"""
Sequence Evaluation
-------------------

Evaluates the variable responses from a sequence in MAD-X.

First: Set all variables to 0
Then: Set one variable at a time to 1

Compare results with case all==0.
"""
import multiprocessing
from contextlib import suppress
from pathlib import Path
from typing import List, Sequence, Tuple

import numpy as np
import tfs

import omc3.madx_wrapper as madx_wrapper
from omc3.correction.response_io import write_fullresponse, write_varmap
from omc3.model.accelerators.accelerator import Accelerator, AccElementTypes
from omc3.utils import logging_tools
from omc3.utils.contexts import timeit

LOG = logging_tools.get_logger(__name__)

EXT = "h5"  # Extension Standard


# Read Sequence ##############################################################


def evaluate_for_variables(
    accel_inst: Accelerator,
    variable_categories,
    order: int = 4,
    num_proc: int = multiprocessing.cpu_count(),
    temp_dir: Path = None
) -> dict:
    """ Generate a dictionary containing response matrices for
        beta, phase, dispersion, tune and coupling and saves it to a file.

        Args:
            accel_inst (Accelerator): Accelerator Instance.
            variable_categories (list): Categories of the variables/knobs to use. (from .json)
            order (int or tuple): Max or [min, max] of K-value order to use.
            num_proc (int): Number of processes to use in parallel.
            temp_dir (Path): temporary directory. If ``None``, uses model_dir.
    """
    LOG.debug("Generating Fullresponse via Mad-X.")
    with timeit(lambda elapsed: LOG.debug(f"  Total time generating fullresponse: {elapsed}s")):
        if not temp_dir:
            temp_dir = Path(accel_inst.model_dir)
        temp_dir.mkdir(parents=True, exist_ok=True)

        variables = accel_inst.get_variables(classes=variable_categories)
        if len(variables) == 0:
            raise ValueError("No variables found! Make sure your categories are valid!")

        num_proc = num_proc if len(variables) > num_proc else len(variables)
        process_pool = multiprocessing.Pool(processes=num_proc)

        k_values = _get_orders(order)

        try:
            _generate_madx_jobs(accel_inst, variables, k_values, num_proc, temp_dir)
            _call_madx(process_pool, temp_dir, num_proc)
            mapping = _load_madx_results(variables, k_values, process_pool, temp_dir)
        finally:
            _clean_up(variables, temp_dir, num_proc)
    return mapping


def _generate_madx_jobs(
    accel_inst: Accelerator,
    variables: Sequence[str],
    k_values: List[float],
    num_proc: int,
    temp_dir: Path,
) -> None:
    """ Generates madx job-files """
    def _assign(var, value):
        return f"{var:s} = {value:d};\n"

    def _do_macro(var):
        return (
            f"exec, create_table(table.{var:s});\n"
            f"write, table=table.{var:s}, file='{str(_get_tablefile(temp_dir, var)):s}';\n"
        )

    LOG.debug("Generating MADX jobfiles.")
    vars_per_proc = int(np.ceil(len(variables) / num_proc))

    # load template
    madx_script: str = _create_basic_job(accel_inst, k_values, variables)

    # build content for testing each variable
    for proc_index in range(num_proc):
        job_content = madx_script % {"TEMPFILE": str(_get_surveyfile(temp_dir, proc_index))}

        for i in range(vars_per_proc):
            try:
                # var to be tested
                current_var = variables[proc_index * vars_per_proc + i]
            except IndexError:
                break
            else:
                job_content += _assign(current_var, 1)
                job_content += _do_macro(current_var)
                job_content += _assign(current_var, 0)
                job_content += "\n"

        # last thing to do: get baseline
        if proc_index+1 == num_proc:
            job_content += _do_macro("0")

        _get_jobfile(temp_dir, proc_index).write_text(job_content)


def _call_madx(process_pool: multiprocessing.Pool, temp_dir: str, num_proc: int) -> None:
    """ Call madx in parallel """
    LOG.debug(f"Starting {num_proc:d} MAD-X jobs...")
    madx_jobs = [_get_jobfile(temp_dir, index) for index in range(num_proc)]
    failed = [LOG.error(fail) for fail in process_pool.map(_launch_single_job, madx_jobs) if fail]
    if len(failed):
        raise RuntimeError(f"{len(failed):d} of {num_proc:d} Madx jobs failed!")
    LOG.debug("MAD-X jobs done.")


def _clean_up(variables: List[str], temp_dir: Path, num_proc: int) -> None:
    """ Merge Logfiles and clean temporary outputfiles """
    LOG.debug("Cleaning output and printing log...")
    for var in (variables + ["0"]):
        table_file = _get_tablefile(temp_dir, var)
        with suppress(FileNotFoundError):  # py3.8 missing_ok=True
            table_file.unlink()
    full_log = ""
    for index in range(num_proc):
        survey_path = _get_surveyfile(temp_dir, index)
        job_path = _get_jobfile(temp_dir, index)
        log_path = job_path.with_name(f"{job_path.name}.log")
        if log_path.exists():
            full_log += log_path.read_text()
        with suppress(FileNotFoundError):  # py3.8 missing_ok=True
            log_path.unlink()

        with suppress(FileNotFoundError):  # py3.8 missing_ok=True
            job_path.unlink()

        with suppress(FileNotFoundError):  # py3.8 missing_ok=True
            survey_path.unlink()
    LOG.debug(full_log)
    with suppress(OSError):
        temp_dir.rmdir()


def _load_madx_results(
    variables: Sequence[str],
    k_values: List[float],
    process_pool: multiprocessing.Pool,
    temp_dir: str,
) -> dict:
    """ Load the madx results in parallel and return var-tfs dictionary """
    LOG.debug("Loading Madx Results.")
    path_and_vars = []
    for value in variables:
        path_and_vars.append((temp_dir, value))

    _, base_tfs = _load_and_remove_twiss((temp_dir, "0"))
    mapping = dict([(order, {}) for order in k_values] + [(f"{order}L", {}) for order in k_values])
    for var, tfs_data in process_pool.map(_load_and_remove_twiss, path_and_vars):
        for order in k_values:
            diff = (tfs_data[order] - base_tfs[order])
            mask = diff != 0  # drop zeros, maybe abs(diff) < eps ?
            k_list = diff.loc[mask]
            mapping[order][var] = k_list
            mapping[f"{order}L"][var] = k_list.mul(base_tfs.loc[mask, "L"])
    return mapping


# Helper #####################################################################


def _get_orders(order: int) -> Sequence[str]:
    """ Returns a list of strings with K-values to be used """
    try:
        return [f"K{i:d}{s:s}" for i in range(3) for s in ["", "S"]]
    except TypeError:
        return [f"K{i:d}{s:s}" for i in range(*order) for s in ["", "S"]]


def _get_jobfile(folder: Path, index: int) -> Path:
    """ Return names for jobfile and iterfile according to index """
    return folder / f"job.varmap.{index:d}.madx"


def _get_tablefile(folder: Path, var: str) -> Path:
    """ Return name of the variable-specific table file """
    return folder / f"table.{var}"


def _get_surveyfile(folder: Path, index: int) -> Path:
    """ Returns the name of the temporary survey file """
    return folder / f"survey.{index:d}.tmp"


def _launch_single_job(inputfile_path: Path):
    """ Function for pool to start a single madx job """
    log_file = inputfile_path.with_name(f"{inputfile_path.name}.log")
    try:
        madx_wrapper.run_file(inputfile_path, log_file=log_file, cwd=inputfile_path.parent)
    except madx_wrapper.MadxError as e:
        return str(e)
    else:
        return None


def _load_and_remove_twiss(path_and_var: Tuple[Path, str]) -> Tuple[str, tfs.TfsDataFrame]:
    """ Function for pool to retrieve results """
    path, var = path_and_var
    twissfile = _get_tablefile(path, var)
    tfs_data = tfs.read(twissfile, index="NAME")
    return var, tfs_data


def _create_basic_job(accel_inst: Accelerator, k_values: List[str], variables: Sequence[str]) -> str:
    """ Create the madx-job basics needed
        TEMPFILE needs to be replaced in the returned string.
    """
    # basic sequence creation
    job_content: str = accel_inst.get_base_madx_script()

    # create a survey and save it to a temporary file
    job_content += (
        "select, flag=survey, clear;\n"
        f"select, flag=survey, pattern='^M.*\.B{accel_inst.beam:d}$', COLUMN=NAME, L;\n"
        "survey, file='%(TEMPFILE)s';\n"
        "readmytable, file='%(TEMPFILE)s', table=mytable;\n"
        "n_elem = table(mytable, tablelength);\n"
        "\n"
    )

    # create macro for assigning values to the k_values per element
    job_content += "assign_k_values(element) : macro = {\n"
    # job_content += "    value, element; show, element;\n"
    for k_val in k_values:
        # (see user guide 10.3 Bending Magnet)
        if k_val == "K0":
            job_content += f"    {k_val:s} = element->angle / element->L;\n"
        elif k_val == "K0S":
            job_content += f"    {k_val:s} = element->tilt / element->L;\n"
        else:
            job_content += f"    {k_val:s} = element->{k_val:s};\n"
    job_content += "};\n\n"

    # create macro for using the row index as variable (see madx userguide)
    job_content += (
        "create_row(tblname, rowidx) : macro = {\n"
        "    exec, assign_k_values(tabstring(tblname, name, rowidx));\n"
        "};\n\n"
    )

    # create macro to create the full table with loop over elements
    job_content += (
        "create_table(table_id) : macro = {\n"
        f"    create, table=table_id, column=_name, L, {','.join(k_values):s};\n"
        "    i_elem = 0;\n"
        "    while (i_elem < n_elem) {\n"
        "        i_elem = i_elem + 1;\n"
        "        setvars, table=mytable, row=i_elem;\n"  # mytable from above!
        "        exec, create_row(mytable, $i_elem);\n"  # mytable from above!
        "        fill,  table=table_id;\n"
        "    };\n"
        "};\n\n"
    )

    # set all variables to zero
    for var in variables:
        job_content += f"{var:s} = 0;\n"

    job_content += "\n"
    return job_content


# Wrapper ##################################################################


def check_varmap_file(accel_inst: Accelerator, vars_categories):
    """ Checks on varmap file and creates it if not in model folder.
    """
    if accel_inst.modifiers is None:
        raise ValueError("Optics not defined. Please provide modifiers.madx. "
                         "Otherwise MADX evaluation might be unstable.")

    varmap_path = Path(accel_inst.model_dir) / f"varmap_{'_'.join(sorted(set(vars_categories)))}.{EXT}"
    if not varmap_path.exists():
        LOG.info(f"Variable mapping '{str(varmap_path):s}' not found. "
                 "Evaluating it via madx.")
        mapping = evaluate_for_variables(accel_inst, vars_categories)
        write_varmap(varmap_path, mapping)

    return varmap_path



"""
Chromatic
----------------

This module contains chromatic calculations functionality of ``optics_measurements``.
It provides functions to compute various chromatic beam properties.
"""
import numpy as np
import pandas as pd

from omc3.optics_measurements.constants import DELTA, ERR, MDL
from omc3.optics_measurements.toolbox import df_prod, df_ratio


def calculate_w_and_phi(betas, dpps, input_files, measure_input, plane):
    columns = [f"{pref}{DELTA}{col}{plane}" for pref in ("", ERR) for col in ("BET", "ALF")]
    joined = betas[0].loc[:, columns]
    for i, beta in enumerate(betas[1:]):
        joined = pd.merge(joined, beta.loc[:, columns], how="inner", left_index=True,
                          right_index=True, suffixes=('', '__' + str(i + 1)))
    for column in columns:
        joined.rename(columns={column: column + '__0'}, inplace=True)
    joined = pd.merge(joined,
                      betas[np.argmin(np.abs(dpps))].loc[:, [f"ALF{plane}", f"{ERR}ALF{plane}"]],
                      how="inner",
                      left_index=True, right_index=True)
    for col in ("BET", "ALF"):
        fit = np.polyfit(np.repeat(dpps, 2),
                         np.repeat(input_files.get_data(joined, f"{DELTA}{col}{plane}").T, 2,
                                   axis=0), 1, cov=True)
        joined[f"D{col}{plane}"] = fit[0][-2, :].T
        joined[f"{ERR}D{col}{plane}"] = np.sqrt(fit[1][-2, -2, :].T)
    a = joined.loc[:, f"DBET{plane}"].to_numpy()
    aerr = joined.loc[:, f"{ERR}DBET{plane}"].to_numpy()
    b = joined.loc[:, f"DALF{plane}"].to_numpy() - joined.loc[:, f"ALF{plane}"].to_numpy() * joined.loc[:,
                                                                                     f"DBET{plane}"].to_numpy()
    berr = np.sqrt(df_prod(joined, f"{ERR}DALF{plane}", f"{ERR}DALF{plane}") +
                   np.square(df_prod(joined, f"{ERR}ALF{plane}", f"DBET{plane}")) +
                   np.square(df_prod(joined, f"ALF{plane}", f"{ERR}DBET{plane}")))
    w = np.sqrt(np.square(a) + np.square(b))
    joined[f"W{plane}"] = w
    joined[f"{ERR}W{plane}"] = np.sqrt(np.square(a * aerr / w) + np.square(b * berr / w))
    joined[f"PHI{plane}"] = np.arctan2(b, a) / (2 * np.pi)
    joined[f"{ERR}PHI{plane}"] = 1 / (1 + np.square(a / b)) * np.sqrt(
        np.square(aerr / b) + np.square(berr * a / np.square(b))) / (2 * np.pi)
    output_df = pd.merge(measure_input.accelerator.model.loc[:,
                         ["S", f"MU{plane}", f"BET{plane}", f"ALF{plane}", f"W{plane}",
                          f"PHI{plane}"]],
                         joined.loc[:,
                         [f"{pref}{col}{plane}" for pref in ("", ERR) for col in ("W", "PHI")]],
                         how="inner", left_index=True,
                         right_index=True, suffixes=(MDL, ''))
    output_df.rename(columns={"SMDL": "S"}, inplace=True)
    return output_df


def calculate_chromatic_coupling(couplings, dpps, input_files, measure_input):
    # TODO how to treat the model values?
    columns = [f"{pref}{col}{part}" for pref in ("", ERR) for col in ("F1001", "F1010") for part in ("RE", "IM")]
    joined = couplings[0].loc[:, columns]
    for i, coup in enumerate(couplings[1:]):
        joined = pd.merge(joined, coup.loc[:, columns], how="inner", left_index=True,
                          right_index=True, suffixes=('', '__' + str(i + 1)))
    for column in columns:
        joined.rename(columns={column: column + '__0'}, inplace=True)

    for col in ("F1001", "F1010"):
        for part in ("RE", "IM"):
            fit = np.polyfit(np.repeat(dpps, 2),
                             np.repeat(input_files.get_data(joined, f"{col}{part}").T, 2,
                                       axis=0), 1, cov=True)
            joined[f"D{col}{part}"] = fit[0][-2, :].T
            joined[f"{ERR}D{col}{part}"] = np.sqrt(fit[1][-2, -2, :].T)
        joined[f"D{col}"] = np.sqrt(np.square(joined.loc[:, f"D{col}RE"].to_numpy()) + np.square(joined.loc[:, f"D{col}IM"].to_numpy()))
        joined[f"{ERR}D{col}"] = np.sqrt(np.square(joined.loc[:, f"D{col}RE"].to_numpy() * df_ratio(joined, f"{ERR}D{col}RE", f"D{col}")) +
                                         np.square(joined.loc[:, f"D{col}IM"].to_numpy() * df_ratio(joined, f"{ERR}D{col}IM", f"D{col}")))
    output_df = pd.merge(measure_input.accelerator.model.loc[:, ["S"]], joined.loc[:,
                         [f"{pref}{col}{part}" for pref in ("", ERR) for col in ("F1001", "F1010") for part in ("", "RE", "IM")]],
                         how="inner", left_index=True,
                         right_index=True)
    return output_df



"""
PETRA
-----

Accelerator-Class for the ``PETRA`` machine.
Model creation not implemented yet.
"""
from omc3.model.accelerators.accelerator import Accelerator, AcceleratorDefinitionError
from omc3.model.constants import PLANE_TO_HV

EXCITER_BPM = "BPM_SOR_13"


class Petra(Accelerator):
    NAME = "petra"

    # Public Methods ##########################################################

    def verify_object(self):  # TODO: Maybe more checks?
        if self.model_dir is None:  # is the class is used to create full response?
            raise AcceleratorDefinitionError("PETRA doesn't have a model creation yet, calling it this "
                                             "way is most probably wrong.")

    def get_exciter_bpm(self, plane, commonbpms):
        return [list(commonbpms).index(EXCITER_BPM), EXCITER_BPM], f"KIFB{PLANE_TO_HV[plane]}N"



"""
PS Model Creator
----------------

This module provides convenience functions for model creation of the ``PS``.
"""
import logging
import shutil

from omc3.model.accelerators.accelerator import AccExcitationMode
from omc3.model.accelerators.ps import Ps
from omc3.model.constants import ERROR_DEFFS_TXT
from omc3.model.model_creators.abstract_model_creator import ModelCreator

LOGGER = logging.getLogger(__name__)


class PsModelCreator(ModelCreator):
    @classmethod
    def get_madx_script(cls, accel: Ps) -> str:
        madx_script = accel.get_base_madx_script()
        replace_dict = {
            "USE_ACD": str(int(accel.excitation == AccExcitationMode.ACD)),
            "DPP": accel.dpp,
            "OUTPUT": str(accel.model_dir),
        }
        madx_template = accel.get_file("twiss.mask").read_text()
        madx_script += madx_template % replace_dict
        return madx_script

    @classmethod
    def get_correction_check_script(cls, accel: Ps, corr_file: str, chrom: bool) -> str:
        raise NotImplemented("Correction check is not implemented for the Ps model creator yet. ")

    @classmethod
    def prepare_run(cls, accel: Ps) -> None:
        # get path of file from PS model directory (without year at the end)
        shutil.copy(accel.get_file("error_deff.txt"), accel.model_dir / ERROR_DEFFS_TXT)



"""
Fitting Tools
-------------

This module contains fitting functionality for ``tune_analysis``.
It provides tools for fitting functions, mainly via odr.

"""
from collections import namedtuple
from typing import Sequence, Dict, List

import numpy as np
import pandas as pd
from numpy.polynomial import Polynomial
from numpy.typing import ArrayLike
from scipy.odr import RealData, Model, ODR
from scipy.optimize import curve_fit

from omc3.utils import logging_tools
from omc3.tune_analysis.constants import FakeOdrOutput, AmpDetData

LOG = logging_tools.get_logger(__name__)


# ODR ###################################################################


def get_poly_fun(order: int):
    """Returns the function of polynomial order. (is this pythonic enough?)."""
    def poly_func(beta, x):
        return sum(beta[i] * np.power(x, i) for i in range(order+1))
    return poly_func


def do_odr(x: pd.Series, y: pd.Series, xerr: pd.Series, yerr: pd.Series, order: int):
    """
    Returns the odr fit.
    Important Convention:
        The beta-parameter in the ODR models go upwards with order, i.e.
        |  beta[0] = y-Axis offset
        |  beta[1] = slope
        |  beta[2] = quadratic term
        |  etc.

    Args:
        x: `Series` of x data.
        y: `Series` of y data.
        xerr: `Series` of x data errors.
        yerr: `Series` of y data errors.
        order: fit order, ``1`` for linear, ``2`` for quadratic.

    Returns: Odr fit. Betas order is index = coefficient of same order.
    """
    LOG.debug("Starting ODR fit.")

    # Poly-Fit for starting point ---
    fit_np = Polynomial.fit(x, y, deg=1).convert()
    LOG.debug(f"ODR fit input (from polynomial fit): {fit_np}")

    # Actual ODR ---
    odr = ODR(data=RealData(x=x, y=y, sx=xerr, sy=yerr),
              model=Model(get_poly_fun(order)),
              beta0=fit_np.coef)
    odr_fit = odr.run()
    logging_tools.odr_pprint(LOG.info, odr_fit)
    return odr_fit


# 2D-Kick ODR ##################################################################

INPUT_ORDER = "qx0", "qy0", "dqx/dex", "dqy/dey",  "dq(x,y)/de(y,x)"


def first_order_detuning_2d(beta: Sequence, x: ArrayLike) -> ArrayLike:
    """ Calculates the 2D tune array (Qx, Qy)
        Qx = qx0 + dqx/dex * ex + dqx/dey * ey
        Qy = qy0 + dqy/dex * ex + dqy/dey * ey

    Args:
        beta: length 5 tune coefficients in order `INPUT_ORDER`
              0: qx0, 1: qy0, 2: xx, 3: yy, 4: xy/yx
        x: array size 2xN, [[ex1, ex2, ...],[ey1, ey2,...]]

    Returns:
        np.array: 2xN [[Qx1, Qx2, ...],[Qy1, Qy2, ...]]
    """
    return np.array([beta[0] + beta[2] * x[0] + beta[4] * x[1],
                     beta[1] + beta[4] * x[0] + beta[3] * x[1]])


def first_order_detuning_2d_jacobian(beta: Sequence, x: ArrayLike) -> ArrayLike:
    """ Jacobian of the 2D tune array:
        [[dqx/dex, dqx/dey ],
         [dqy/dex, dqy/dey ]]

    Args:
        beta: length 5 tune coefficients in order `INPUT_ORDER`
              0: qx0, 1: qy0, 2: xx, 3: yy, 4: xy/yx
        x: length 2 Sequence, (ex, ey)

    Returns:
        np.array: size 2x2xlen(x) containing the Jacobian of the 2d-fit function.
    """
    return np.dstack([np.array([[beta[2], beta[4]],
                     [beta[4], beta[3]]])] * len(x[0]))


def map_odr_fit_to_planes(odr_fit) -> Dict[str, Dict[str, FakeOdrOutput]]:
    """ Maps the calculated odr fit to fake odr fits with `beta` and `sd_beta`
    attributes. These would be the results of first-order amplitude detuning
    odr-fits when done independently by tune and kick plane.

    Returns: Dict[str, Dict[str: odr_fit]] of ODR fits, where the inner
             string gives the kick-plane, the outer the tune-plane..

    """
    def get_fit(a: int, b: int):
        return FakeOdrOutput(
            beta=[odr_fit.beta[a], odr_fit.beta[b]],
            sd_beta=[odr_fit.sd_beta[a], odr_fit.sd_beta[b]]
        )

    return {
        "X": {
            "X": get_fit(0, 2),
            "Y": get_fit(0, 4),
        },
        "Y": {
            "X": get_fit(1, 4),
            "Y": get_fit(1, 3),
        },
    }


def do_2d_kicks_odr(x: ArrayLike, y: ArrayLike, xerr: ArrayLike, yerr: ArrayLike):
    """
    Returns the odr fit.

    Args:
        x: `Array` of x data (2xN).
        y: `Array` of y data (2xN).
        xerr: `Array` of x data errors (2xN).
        yerr: `Array` of y data errors (2xN).

    Returns: Dict[str, Dict[str: odr_fit]] of Odr fits, where the inner
             string gives the kick-plane, the outer the tune-plane..
    """
    LOG.debug("Starting ODR fit.")

    x, y, xerr, yerr = _filter_nans(x, y, xerr, yerr)

    # Curve-Fit for starting point ---
    curve_fit_fun = lambda v, *args: first_order_detuning_2d(args, v).ravel()
    beta, beta_cov = curve_fit(f=curve_fit_fun, xdata=x, ydata=y.ravel(), p0=[0]*5)

    res_str = ",\n".join([f"{n:>16} = {b:9.3g}" for n, b in zip(INPUT_ORDER, beta)])
    LOG.info(f"\nDetuning estimate without errors (curve fit):\n{res_str}\n")

    # Actual ODR ---
    odr = ODR(data=RealData(x=x, y=y, sx=xerr, sy=yerr),
              # model=Model(first_order_detuning_2d),
              model=Model(first_order_detuning_2d, fjacd=first_order_detuning_2d_jacobian),
              beta0=beta)
    odr_fit = odr.run()
    logging_tools.odr_pprint(LOG.debug, odr_fit)

    res_str = ",\n".join([f"{n:>16} = {b:9.3g} +- {e:8.3g}" for n, b, e in zip(INPUT_ORDER, odr_fit.beta, odr_fit.sd_beta)])
    LOG.info(f"\nDetuning estimate with errors (odr):\n{res_str}\n")
    return map_odr_fit_to_planes(odr_fit)


def _filter_nans(*args: ArrayLike) -> List[ArrayLike]:
    """Remove all data points containing a NaN.
    Assumes input arrays are all of shape 2xN.
    TODO:
    As this is not done in plotting, points might be plotted,
    that have not been used for fitting.
    """
    a = np.array(args)
    a = a[:, :, ~np.isnan(a).any(axis=0).any(axis=0)]
    return list(a)




"""
Model Appenders
---------------

Utilities to append new columns to measurement and model dataframes.
E.g. get differences between measurement and model and append those to
the measurement data (for corrections).
"""
from collections import defaultdict
from typing import Callable, Dict, Sequence

import numpy as np
import pandas as pd
from optics_functions.coupling import coupling_via_cmatrix

from omc3.correction.constants import (BETA, DIFF, MODEL, NORM_DISP, PHASE_ADV,
                                       TUNE, VALUE, F1001, F1010, PHASE)
from omc3.utils import logging_tools
from omc3.optics_measurements.toolbox import df_diff, df_rel_diff

LOG = logging_tools.get_logger(__name__)


def add_differences_to_model_to_measurements(
    model: pd.DataFrame,
    measurement: Dict[str, pd.DataFrame],
    keys: Sequence[str] = None,
) -> Dict[str, pd.DataFrame]:
    """
    Provided with DataFrames from a model and a measurement, and a number of keys to be found in both,
    returns a dictionary with the variation from measurement to model for each key.

    Args:
        model (pd.DataFrame): DataFrame of the model.
        measurement (Dict[str, pd.DataFrame]): DataFrames of the measurement.
        keys (Sequence[str]): Parameters to get variation to model for. Optional.
                              If omitted, all entries in `measurement` are used.

    Returns:
        A dictionary of optics parameters and the resulting DataFrames.
    """
    appenders: Dict[str, Callable] = _get_model_appenders()
    if keys is None:
        keys = measurement.keys()

    res_dict = {}

    for key in keys:
        res_dict[key] = appenders[key](model, measurement[key].copy(), key)
    return res_dict


def _get_model_appenders() -> Dict[str, Callable]:
    return defaultdict(lambda:  _get_model_generic, {
        f"{PHASE}X": _get_model_phases, f"{PHASE}Y": _get_model_phases,
        f"{BETA}X": _get_model_betabeat, f"{BETA}Y": _get_model_betabeat,
        f"{NORM_DISP}X": _get_model_norm_disp, f"{TUNE}": _get_model_tunes, })


def _get_model_generic(model: pd.DataFrame, meas: pd.DataFrame, key: str) -> pd.DataFrame:
    with logging_tools.log_pandas_settings_with_copy(LOG.debug):
        meas[MODEL] = model.loc[meas.index.to_numpy(), key].to_numpy()
        meas[DIFF] = df_diff(meas, VALUE, MODEL)
    return meas


def _get_model_phases(model: pd.DataFrame, meas: pd.DataFrame, key: str) -> pd.DataFrame:
    model_column = f"{PHASE_ADV}{key[-1]}"
    with logging_tools.log_pandas_settings_with_copy(LOG.debug):
        meas[MODEL] = (model.loc[meas["NAME2"].to_numpy(), model_column].to_numpy() -
                       model.loc[meas.index.to_numpy(), model_column].to_numpy())
        meas[DIFF] = df_diff(meas, VALUE, MODEL)
    return meas


def _get_model_betabeat(model: pd.DataFrame, meas: pd.DataFrame, key: str) -> pd.DataFrame:
    with logging_tools.log_pandas_settings_with_copy(LOG.debug):
        meas[MODEL] = model.loc[meas.index.to_numpy(), key].to_numpy()
        meas[DIFF] = df_rel_diff(meas, VALUE, MODEL)
    return meas


def _get_model_norm_disp(model: pd.DataFrame, meas: pd.DataFrame, key: str) -> pd.DataFrame:
    col = key[1:]
    beta = f"{BETA}{key[-1]}"
    with logging_tools.log_pandas_settings_with_copy(LOG.debug):
        meas[MODEL] = model.loc[meas.index.to_numpy(), col].to_numpy() / np.sqrt(model.loc[meas.index.to_numpy(), beta].to_numpy())
        meas[DIFF] = meas.loc[:, VALUE].to_numpy() - meas.loc[:, MODEL].to_numpy()
    return meas


def _get_model_tunes(model: pd.DataFrame, meas: pd.DataFrame, key: str) -> pd.DataFrame:
    # We want just fractional tunes
    with logging_tools.log_pandas_settings_with_copy(LOG.debug):
        meas[MODEL] = np.remainder([model[f"{TUNE}1"], model[f"{TUNE}2"]], [1, 1])
        meas[DIFF] = meas.loc[:, VALUE].to_numpy() - meas.loc[:, MODEL].to_numpy()
    return meas


# ----

def add_coupling_to_model(model: pd.DataFrame) -> pd.DataFrame:
    """
    Computes the coupling RDTs from the input model TfsDataFrame and returns a copy of said TfsDataFrame with
    columns for the real and imaginary parts of the computed coupling RDTs.

    Args:
        model (tfs.TfsDataFrame): Twiss dataframe.

    Returns:
        A TfsDataFrame with the added columns.
    """
    result_tfs_df = model.copy()
    coupling_rdts_df = coupling_via_cmatrix(result_tfs_df)
    result_tfs_df[f"{F1001}R"] = np.real(coupling_rdts_df[f"{F1001}"]).astype(np.float64)
    result_tfs_df[f"{F1001}I"] = np.imag(coupling_rdts_df[f"{F1001}"]).astype(np.float64)
    result_tfs_df[f"{F1010}R"] = np.real(coupling_rdts_df[f"{F1010}"]).astype(np.float64)
    result_tfs_df[f"{F1010}I"] = np.imag(coupling_rdts_df[f"{F1010}"]).astype(np.float64)
    return result_tfs_df




"""
Dispersion
----------

This module contains dispersion calculations related functionality of ``optics_measurements``.
It provides functions to compute orbit, dispersion and normalised dispersion.
"""
from os.path import join

import numpy as np
import pandas as pd
import tfs

from omc3.definitions.constants import PI2I
from omc3.optics_measurements.constants import (DELTA, DISPERSION_NAME, ERR,
                                                EXT, MDL, NORM_DISP_NAME,
                                                ORBIT_NAME)
from omc3.utils import stats


def calculate_orbit(meas_input, input_files, header, plane):
    """
    Calculates orbit.

    Args:
        meas_input: `OpticsInput` object
        input_files: Stores the input files tfs.
        header: `OrderedDict` containing information about the analysis.
        plane: marking the horizontal or vertical plane, **X** or **Y**.

    Returns:
        `TfsDataFrame` corresponding to output file.
    """
    df_orbit = _get_merged_df(meas_input, input_files, plane, ['CO', 'CORMS'])
    df_orbit[plane] = stats.weighted_mean(input_files.get_data(df_orbit, 'CO'), axis=1)
    df_orbit[f"{ERR}{plane}"] = stats.weighted_error(input_files.get_data(df_orbit, 'CO'), axis=1)
    df_orbit = _get_delta_columns(df_orbit, plane)
    output_df = df_orbit.loc[:, _get_output_columns(plane, df_orbit)]
    tfs.write(join(meas_input.outputdir, f"{ORBIT_NAME}{plane.lower()}{EXT}"), output_df, header, save_index='NAME')
    return output_df


def calculate_dispersion(meas_input, input_files, header_dict, plane):
    """
    Calculates dispersion.

    Args:
        meas_input: `OpticsInput` object.
        input_files: Stores the input files tfs.
        header_dict: `OrderedDict` containing information about the analysis.
        plane: marking the horizontal or vertical plane, **X** or **Y**.

    Returns:
        `TfsDataFrame` corresponding to output file.
    """
    if meas_input.three_d_excitation:
        return _calculate_dispersion_3d(meas_input, input_files, header_dict, plane)
    return _calculate_dispersion_2d(meas_input, input_files, header_dict, plane)


def calculate_normalised_dispersion(meas_input, input_files, beta, header_dict):
    """
    Calculates normalised dispersion.

    Args:
        meas_input: `OpticsInput` object.
        input_files: Stores the input files tfs.
        beta: measured betas to get dispersion from normalised dispersion.
        header_dict: `OrderedDict` containing information about the analysis.

    Returns:
        `TfsDataFrame` corresponding to output file.
    """
    if meas_input.three_d_excitation:
        return _calculate_normalised_dispersion_3d(meas_input, input_files, beta, header_dict)
    return _calculate_normalised_dispersion_2d(meas_input, input_files, beta, header_dict)


def _calculate_dispersion_2d(meas_input, input_files, header, plane):
    order = 2 if meas_input.second_order_dispersion else 1
    dpps = input_files.dpps(plane)
    if np.max(dpps) - np.min(dpps) == 0.0:
        return  # temporary solution
    model = meas_input.accelerator.model
    df_orbit = _get_merged_df(meas_input, input_files, plane, ['CO', 'CORMS'])
    fit = np.polyfit(dpps, input_files.get_data(df_orbit, 'CO').T, order, cov=True)
    # in the fit results the coefficients are sorted by power in decreasing order
    if order > 1:
        df_orbit[f"D2{plane}"] = fit[0][-3, :].T
        df_orbit[f"{ERR}D2{plane}"] = np.sqrt(fit[1][-3, -3, :].T)
    df_orbit[f"D{plane}"] = fit[0][-2, :].T
    df_orbit[f"{ERR}D{plane}"] = np.sqrt(fit[1][-2, -2, :].T)
    df_orbit[plane] = fit[0][-1, :].T
    df_orbit[f"{ERR}{plane}"] = np.sqrt(fit[1][-1, -1, :].T)
    # since we get variances from the fit, maybe we can include the variances of fitted points
    df_orbit[f"DP{plane}"] = _calculate_dp(model,
                                           df_orbit.loc[:, [f"D{plane}", f"{ERR}D{plane}"]], plane)
    df_orbit = _get_delta_columns(df_orbit, plane)
    output_df = df_orbit.loc[:, _get_output_columns(plane, df_orbit)]
    tfs.write(join(meas_input.outputdir, f"{DISPERSION_NAME}{plane.lower()}{EXT}"), output_df, header, save_index='NAME')
    return output_df


def _calculate_dispersion_3d(meas_input, input_files, header_dict, plane):
    """Computes dispersion from 3D kicks."""
    output, accelerator = meas_input.outputdir, meas_input.accelerator
    model = accelerator.model
    df_orbit = _get_merged_df(meas_input, input_files, plane, ['AMPZ', 'MUZ', f"AMP{plane}"])
    # work around due to scaling to main line in lin files
    unscaled_amps = (df_orbit.loc[:, input_files.get_columns(df_orbit, 'AMPZ')].to_numpy() *
                     df_orbit.loc[:, input_files.get_columns(df_orbit, f"AMP{plane}")].to_numpy())
    mask = accelerator.get_element_types_mask(df_orbit.index, ["arc_bpm"])
    global_factors = np.array([1 / df.headers["DPPAMP"] for df in input_files[plane]])
    # scaling to the model, and getting the synchrotron phase in the arcs
    df_orbit[f"D{plane}"], df_orbit[f"{ERR}D{plane}"] = _get_signed_dispersion(
            input_files, df_orbit, unscaled_amps * global_factors, mask)
    df_orbit[f"DP{plane}"] = _calculate_dp(model, df_orbit.loc[:, [f"D{plane}", f"{ERR}D{plane}"]], plane)
    df_orbit = _get_delta_columns(df_orbit, plane)
    output_df = df_orbit.loc[:, _get_output_columns(plane, df_orbit)]
    tfs.write(join(output, f"{DISPERSION_NAME}{plane.lower()}{EXT}"), output_df, header_dict, save_index='NAME')
    return output_df


def _calculate_normalised_dispersion_2d(meas_input, input_files, beta, header):
    # TODO there are no errors from orbit
    order = 2 if meas_input.second_order_dispersion else 1
    plane = "X"
    model = meas_input.accelerator.model
    df_orbit = _get_merged_df(meas_input, input_files, plane, ['CO', 'CORMS', f"AMP{plane}"])
    df_orbit[f"ND{plane}{MDL}"] = df_orbit.loc[:, f"D{plane}{MDL}"] / np.sqrt(
        df_orbit.loc[:, f"BET{plane}{MDL}"])
    if order > 1:
        df_orbit[f"ND2{plane}{MDL}"] = df_orbit.loc[:, f"D2{plane}{MDL}"] / np.sqrt(
            df_orbit.loc[:, f"BET{plane}{MDL}"])
    df_orbit = pd.merge(df_orbit, beta.loc[:, [f"BET{plane}", f"{ERR}BET{plane}"]], how='inner',
                        left_index=True, right_index=True)
    dpps = input_files.dpps(plane)
    if np.max(dpps) - np.min(dpps) == 0.0:
        return  # temporary solution
        # raise ValueError('Cannot calculate dispersion, only a single dpoverp')
    fit = np.polyfit(dpps, input_files.get_data(df_orbit, 'CO').T, order, cov=True)
    if order > 1:
        df_orbit['ND2X_unscaled'] = fit[0][-3, :].T / stats.weighted_mean(input_files.get_data(df_orbit, f"AMP{plane}"), axis=1)
        df_orbit['STDND2X_unscaled'] = np.sqrt(fit[1][-3, -3, :].T) / stats.weighted_mean(input_files.get_data(df_orbit, f"AMP{plane}"), axis=1)
    df_orbit['NDX_unscaled'] = fit[0][-2, :].T / stats.weighted_mean(input_files.get_data(df_orbit, f"AMP{plane}"), axis=1)  # TODO there is no error from AMPX
    df_orbit['STDNDX_unscaled'] = np.sqrt(fit[1][-2, -2, :].T) / stats.weighted_mean(input_files.get_data(df_orbit, f"AMP{plane}"), axis=1)
    mask = meas_input.accelerator.get_element_types_mask(df_orbit.index, ["arc_bpm"])
    global_factor = np.sum(df_orbit.loc[mask, f"ND{plane}{MDL}"].to_numpy()) / np.sum(df_orbit.loc[mask, 'NDX_unscaled'].to_numpy())
    if order > 1:
        df_orbit[f"ND2{plane}"] = global_factor * df_orbit.loc[:, 'ND2X_unscaled']
        df_orbit[f"{ERR}ND2{plane}"] = global_factor * df_orbit.loc[:, 'STDND2X_unscaled']
    df_orbit[f"ND{plane}"] = global_factor * df_orbit.loc[:, 'NDX_unscaled']
    df_orbit[f"{ERR}ND{plane}"] = global_factor * df_orbit.loc[:, 'STDNDX_unscaled']
    df_orbit = _calculate_from_norm_disp(df_orbit, model, plane)
    output_df = df_orbit.loc[:, _get_output_columns(plane, df_orbit)]
    tfs.write(join(meas_input.outputdir, f"{NORM_DISP_NAME}{plane.lower()}{EXT}"), output_df, header, save_index='NAME')
    return output_df


def _calculate_normalised_dispersion_3d(meas_input, input_files, beta, header):
    """
    Computes horizontal normalised dispersion from 3D kicks, and performs model-based
    compensation, i.e. as in _free2 files.
    """
    output, accelerator = meas_input.outputdir, meas_input.accelerator
    model = accelerator.model
    driven_model = accelerator.model_driven if accelerator.excitation else model
    plane = "X"
    df_orbit = _get_merged_df(meas_input, input_files, plane, ['AMPZ', 'MUZ'])
    df_orbit[f"ND{plane}{MDL}"] = df_orbit.loc[:, f"D{plane}{MDL}"] / np.sqrt(df_orbit.loc[:, f"BET{plane}{MDL}"])
    df_orbit = pd.merge(df_orbit, beta.loc[:, [f"BET{plane}", f"{ERR}BET{plane}"]], how='inner', left_index=True, right_index=True)
    df_orbit = pd.merge(df_orbit, driven_model.loc[:, [f"BET{plane}"]], how='inner', left_index=True,
                        right_index=True, suffixes=('', '_driven'))
    mask = accelerator.get_element_types_mask(df_orbit.index, ["arc_bpm"])
    compensation = np.sqrt(df_orbit.loc[:, f"BET{plane}_driven"].to_numpy() / df_orbit.loc[:, f"BET{plane}{MDL}"].to_numpy())
    global_factors = np.sum(df_orbit.loc[mask, f"ND{plane}{MDL}"].to_numpy()) / np.sum(df_orbit.loc[mask, input_files.get_columns(df_orbit, 'AMPZ')].to_numpy() * compensation[mask, None], axis=0)
    # scaling to the model, and getting the synchrotron phase in the arcs
    scaled_amps = (df_orbit.loc[:, input_files.get_columns(df_orbit, 'AMPZ')].to_numpy() * global_factors) * compensation[:, None]
    df_orbit[f"ND{plane}"], df_orbit[f"{ERR}ND{plane}"] = _get_signed_dispersion(
        input_files, df_orbit, scaled_amps, mask)
    df_orbit = _calculate_from_norm_disp(df_orbit, model, plane)
    output_df = df_orbit.loc[:, _get_output_columns(plane, df_orbit)]
    tfs.write(join(output, f"{NORM_DISP_NAME}{plane.lower()}{EXT}"), output_df, header, save_index='NAME')
    return output_df


def _calculate_dp(model, disp, plane):
    _m = "meas"
    df = pd.DataFrame(model).loc[:, ['S', f"MU{plane}", f"DP{plane}", f"D{plane}",
                                     f"BET{plane}", f"ALF{plane}"]]
    df = pd.merge(df, disp.loc[:, [f"D{plane}", f"{ERR}D{plane}"]], how='inner', left_index=True,
                  right_index=True, suffixes=('', _m))
    shifted = np.roll(df.index.to_numpy(), -1)
    p_mdl_12 = df.loc[shifted, f"MU{plane}"].to_numpy() - df.loc[:, f"MU{plane}"].to_numpy()
    p_mdl_12[-1] = p_mdl_12[-1] + model['Q' + str(1+(plane == "Y"))]
    phi_12 = p_mdl_12 * 2 * np.pi
    m11 = np.sqrt(df.loc[shifted, f"BET{plane}"] / df.loc[:, f"BET{plane}"]
                  ) * (np.cos(phi_12) + df.loc[:, f"ALF{plane}"] * np.sin(phi_12))
    m12 = np.sqrt(df.loc[shifted, f"BET{plane}"] * df.loc[:, f"BET{plane}"]) * np.sin(phi_12)
    m13 = df.loc[shifted, f"D{plane}"] - m11 * df.loc[:, f"D{plane}"] - m12 * df.loc[:, f"DP{plane}"]
    return (-m13 + df.loc[shifted, f"D{plane}{_m}"] - m11 * df.loc[:, f"D{plane}{_m}"]) / m12


def _get_merged_df(meas_input, input_files, plane, meas_columns):
    model = meas_input.accelerator.model
    df = pd.DataFrame(model).reindex(columns=["S", plane, f"D{plane}", f"DP{plane}", f"MU{plane}",
                                              f"BET{plane}", f"DD{plane}"], fill_value=np.nan)
    df.rename(columns={plane: f"{plane}{MDL}", f"D{plane}": f"D{plane}{MDL}", f"DP{plane}": f"DP{plane}{MDL}",
                       f"MU{plane}": f"MU{plane}{MDL}", f"BET{plane}": f"BET{plane}{MDL}", f"DD{plane}": f"D2{plane}{MDL}"}, inplace=True)
    if not meas_input.second_order_dispersion:
        df.drop(columns=[f'D2{plane}{MDL}'], inplace=True)
    df = pd.merge(df, input_files.joined_frame(plane, meas_columns, dpp_amp=meas_input.three_d_excitation), how='inner', left_index=True, right_index=True)
    df['COUNT'] = len(input_files.get_columns(df, meas_columns[0]))
    return df


def _get_signed_dispersion(input_files, df_orbit, scaled_amps, mask):
    same_interval_phase = np.angle(np.exp(PI2I * df_orbit.loc[:, input_files.get_columns(df_orbit, 'MUZ')].to_numpy())) / (2 * np.pi)
    phase_wrt_arcs = same_interval_phase - stats.circular_mean(same_interval_phase[mask, :], period=1, axis=0)
    phase_wrt_arcs = np.abs(np.where(np.abs(phase_wrt_arcs) > 0.5, phase_wrt_arcs - np.sign(phase_wrt_arcs), phase_wrt_arcs))
    if len(input_files.get_columns(df_orbit, 'AMPZ')) > 1:
        # resolving the sign of dispersion
        dispersions = scaled_amps * np.sign(0.25 - np.abs(stats.circular_mean(phase_wrt_arcs, period=1, axis=1)))[:, None]
        # final calculation
        return np.mean(dispersions, axis=1), np.std(dispersions, axis=1) * stats.t_value_correction(
            dispersions.shape[1])
    return scaled_amps * np.sign(0.25 - np.abs(phase_wrt_arcs)), 0.0


def _get_output_columns(plane, df):
    cols = (["S", "COUNT", f"MU{plane}MDL"] +           # common columns
            _single_column_set_list(plane) +            # orbit columns
            _single_column_set_list(f"ND{plane}") +     # normalized dispersion columns
            _single_column_set_list(f"D{plane}") +      # dispersion columns
            [f"DP{plane}", f"DP{plane}{MDL}"] +         # more dispersion columns
            _single_column_set_list(f"D2{plane}") +     # second order dispersion columns
            _single_column_set_list(f"ND2{plane}"))     # second order normalized dispersion columns
    return [col for col in cols if col in df.columns]


def _single_column_set_list(base_name):
    return [f"{base_name}", f"{ERR}{base_name}", f"{DELTA}{base_name}", f"{ERR}{DELTA}{base_name}", f"{base_name}{MDL}"]


def _calculate_from_norm_disp(df, model, plane):
    df[f"D{plane}"] = df.loc[:, f"ND{plane}"] * np.sqrt(df.loc[:, f"BET{plane}"])
    df[f"{ERR}D{plane}"] = df.loc[:, f"{ERR}ND{plane}"] * np.sqrt(df.loc[:, f"BET{plane}"])
    df[f"DP{plane}"] = _calculate_dp(model, df.loc[:, [f"D{plane}", f"{ERR}D{plane}"]], plane)
    return _get_delta_columns(df, plane)


def _get_delta_columns(df, plane):
    for col in [f"{plane}", f"D{plane}", f"ND{plane}", f"D2{plane}", f"ND2{plane}"]:
        if col in df.columns:
            df[f"{DELTA}{col}"] = df.loc[:, col] - df.loc[:, f"{col}{MDL}"]
            df[f"{ERR}{DELTA}{col}"] = df.loc[:, f"{ERR}{col}"]
    return df



"""
Model Diff
----------

Calculate the differences in optics parameters between twiss-models.
Similar to :func:`omc3.correction.model_appenders.add_differences_to_model_to_measurements`,
yet operates on two twiss files instead.

"""
from collections import defaultdict
from typing import Sequence

import tfs

from omc3.correction.constants import DISP, BETA, PHASE_ADV, TUNE, PHASE, NORM_DISP, DELTA
from omc3.optics_measurements.toolbox import ang_diff
import numpy as np


def diff_twiss_parameters(model_a: tfs.TfsDataFrame,
                          model_b: tfs.TfsDataFrame,
                          parameters: Sequence[str] = None) -> tfs.TfsDataFrame:
    """Create a TfsDataFrame containing of the given parameters between
    model_a and model_b."""
    # preparation ---
    if parameters is None:
        parameters = model_a.columns.intersection(model_b.columns)

    index = model_a.index.intersection(model_b.index)
    model_a, model_b = model_a.loc[index, :], model_b.loc[index, :]

    # get diff dataframe ---
    with_tune = TUNE in parameters
    if with_tune:
        parameters = [p for p in parameters if p != TUNE]

    diff_df = tfs.TfsDataFrame(index=index, columns=[f"{DELTA}{p}" for p in parameters])
    fun_map = _get_mapping()
    for parameter in parameters:
        diff_function = fun_map[parameter[:-1]]
        diff_df.loc[:, f"{DELTA}{parameter}"] = diff_function(model_a, model_b, parameter)

    if with_tune:
        for tune in (f"{TUNE}1", f"{TUNE}2"):
            diff_df.headers[f"{DELTA}{tune}"] = _tune_diff(model_a, model_b, tune)

    return diff_df


# Helper -----------------------------------------------------------------------


def _get_mapping():
    return defaultdict(
        lambda: _diff,
        {
            PHASE_ADV: _phase_advance_diff,
            PHASE: _phase_advance_diff,
            BETA: _beta_beating,
            NORM_DISP: _normalized_dispersion_diff,
        }
    )


def _diff(model_a, model_b, column):
    return model_a.loc[:, column] - model_b.loc[:, column]


def _tune_diff(model_a, model_b, column):
    return model_a.headers[column] % 1 - model_b.headers[column] % 1


def _phase_advance_diff(model_a, model_b, column):
    column = f"{PHASE_ADV}{column[-1]}"  # in case it's "PHASE", which is a parameter of the measurement
    name = model_a.index[:-1]
    name2 = model_a.index[1:]
    return np.append(
        [0],
        ang_diff(
            ang_diff(model_a.loc[name2, column].to_numpy(), model_a.loc[name, column].to_numpy()),
            ang_diff(model_b.loc[name2, column].to_numpy(), model_b.loc[name, column].to_numpy())
        )
    )


def _beta_beating(model_a, model_b, column):
    return np.divide(model_a.loc[:, column], model_b.loc[:, column]) - 1


def _normalized_dispersion_diff(model_a, model_b, column):
    plane = column[-1]
    return -np.diff([
        np.divide(model.loc[:, f"{DISP}{plane}"], np.sqrt(model.loc[:, f"{BETA}{plane}"])).to_numpy()
        for model in (model_a, model_b)
    ], axis=0).T



"""
Tune
----
This module contains tune calculations functionality of ``optics_measurements``.
It provides functions to compute betatron tunes and structures to store them.
"""
import numpy as np
import pandas as pd

from omc3.definitions.constants import PLANES, PLANE_TO_NUM
from omc3.utils import stats


def calculate(measure_input, input_files):
    tune_d = TuneDict()
    accelerator = measure_input.accelerator
    for plane in PLANES:
        tune_d[plane]["QM"] = accelerator.model.headers[f"Q{PLANE_TO_NUM[plane]}"]
        tune_list = [df.headers[f"Q{PLANE_TO_NUM[plane]}"] for df in input_files.dpp_frames(plane, 0)]
        tune_rms_list = [df.headers[f"Q{PLANE_TO_NUM[plane]}RMS"] for df in input_files.dpp_frames(plane, 0)]
        measured_tune = stats.weighted_mean(np.array(tune_list), errors=np.array(tune_rms_list))
        tune_d[plane]["Q"], tune_d[plane]["QF"] = measured_tune, measured_tune
        tune_d[plane]["QFM"] = accelerator.nat_tunes[PLANE_TO_NUM[plane] - 1]
        if accelerator.excitation:
            tune_d[plane]["QM"] = accelerator.drv_tunes[PLANE_TO_NUM[plane] - 1]
            tune_d[plane]["QF"] = tune_d[plane]["Q"] - tune_d[plane]["QM"] + tune_d[plane]["QFM"]
            if measure_input.compensation == "equation":
                tune_d[plane]["ac2bpm"] = tune_d.phase_ac2bpm(
                    input_files.joined_frame(plane, [f"MU{plane}"], dpp_value=0, how='inner'),
                    plane, measure_input.accelerator)
    return tune_d


class TuneDict(dict):
    """
    Data structure to hold tunes.
    """
    # TODO: detail each key
    def __init__(self):
        super(TuneDict, self).__init__(zip(PLANES, ({"Q": 0.0, "QF": 0.0, "QM": 0.0, "QFM": 0.0, "ac2bpm": None},
                                                    {"Q": 0.0, "QF": 0.0, "QM": 0.0, "QFM": 0.0, "ac2bpm": None})))

    def get_lambda(self, plane):
        """
        Computes lambda compensation factor.

        Args:
            plane: marking the horizontal or vertical plane, **X** or **Y**.

        Returns:
             lambda compensation factor (driven vs free motion).
        """
        return (np.sin(np.pi * (self[plane]["Q"] - self[plane]["QF"])) /
                np.sin(np.pi * (self[plane]["Q"] + self[plane]["QF"])))

    def phase_ac2bpm(self, df_idx_by_bpms: pd.DataFrame, plane: str, accelerator):
        """
        Returns the necessary values for the exciter compensation.
        See **DOI: 10.1103/PhysRevSTAB.11.084002**

        Args:
            df_idx_by_bpms (pandas.DataFrame): commonbpms (see GetLLM._get_commonbpms)
            plane (str): marking the horizontal or vertical plane, **X** or **Y**.
            accelerator: an `Accelerator` object.

        Returns:
            A `Tuple` consisting of four elements a, b, c, d.
                - a (string): name of the nearest BPM.
                - b (float): compensated phase advance between the exciter and the nearest BPM.
                - c (int): k of the nearest BPM.
                - d (string): name of the exciter element.
        """
        model = accelerator.elements
        r = self.get_lambda(plane)
        [k, bpmac1], exciter = accelerator.get_exciter_bpm(plane, df_idx_by_bpms.index)
        psi = model.loc[bpmac1, f"MU{plane}"] - model.loc[exciter, f"MU{plane}"]
        psi = np.arctan((1 + r) / (1 - r) * np.tan(
            2 * np.pi * psi + np.pi * self[plane]["QF"])) % np.pi - np.pi * self[plane]["Q"]
        psi = psi / (2 * np.pi)
        return bpmac1, psi, k, exciter



"""
Abstract Model Creator Class
----------------------------

This module provides the template for all model creators.
"""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Sequence, Union

from omc3.model.accelerators.accelerator import Accelerator, AccExcitationMode
from omc3.model.constants import TWISS_AC_DAT, TWISS_ADT_DAT, TWISS_DAT, TWISS_ELEMENTS_DAT


class ModelCreator(ABC):
    """
    Abstract class for the implementation of a model creator. All mandatory methods and convenience
    functions are defined here.
    """
    @classmethod
    @abstractmethod
    def get_correction_check_script(cls, accel: Accelerator, corr_file: str, chrom: bool) -> str:
        """
        Returns the ``MAD-X`` script used to verify global corrections. This script should create twiss
        files for before (``twiss_no.dat``) and after (``twiss_corr.dat``) correction.

        Args:
            accel (Accelerator): Accelerator Instance used for the model creation.
            corr_file (str): File containing the corrections (madx-readable).
            chrom (bool): Flag for chromatic corrections deltapm and deltapp.

        Returns:
            The string of the ``MAD-X`` script used to verify global corrections.
        """
        pass

    @classmethod
    @abstractmethod
    def get_madx_script(cls, accel: Accelerator) -> str:
        """
        Returns the ``MAD-X`` script used to create the model (directory).

        Args:
            accel (Accelerator): Accelerator Instance used for the model creation.

        Returns:
            The string of the ``MAD-X`` script used to used to create the model.
        """
        pass

    @classmethod
    @abstractmethod
    def prepare_run(cls, accel: Accelerator) -> None:
        """
        Prepares the model creation ``MAD-X`` run. It should check that the appropriate directories
        are created, and that macros and other files are in place.
        Should also check that all necessary data for model creation is available in the accelerator
        instance. Called by the ``model_creator.create_accel_and_instance``

        Args:
            accel (Accelerator): Accelerator Instance used for the model creation.
        """
        pass

    @classmethod
    def check_run_output(cls, accel: Accelerator) -> None:
        """
        Checks that the model creation ``MAD-X`` run was successful. It should check that the
        appropriate directories are created, and that macros and other files are in place.
        Checks the accelerator instance. Called by the ``model_creator.create_instance_and_model``

        Args:
            accel (Accelerator): Accelerator Instance used for the model creation.
        """
        # These are the default files for most model creators for now.
        files_to_check: List[str] = [TWISS_DAT, TWISS_ELEMENTS_DAT]
        if accel.excitation == AccExcitationMode.ACD:
            files_to_check += [TWISS_AC_DAT]
        elif accel.excitation == AccExcitationMode.ADT:
            files_to_check += [TWISS_ADT_DAT]
        cls._check_files_exist(accel.model_dir, files_to_check)

    @staticmethod
    def _check_files_exist(dir_: Union[Path, str], files: Sequence[str]) -> None:
        """
        Convenience function to loop over files supposed to be in a locatioin and raise an error if
        one or more of these files does not exist.

        Args:
            dir_ (Union[Path, str]): Path object or string of the absolute path of the directory in
                which to check for the files.
            files (Sequence[str]): the names of files to check the presence of in the directory.

        Raises:
            Raises an ``FileNotFoundError``
        """
        for out_file in files:
            file_path = Path(dir_) / out_file
            if not file_path.exists():
                raise FileNotFoundError(
                    f"Model Creation Failed. The file '{file_path.absolute()}' was not created."
                )



"""
Plot Spectrum - Stem Plotter
-----------------------------

Stem plotting functionality for spectrum plotter.
"""
import matplotlib
from generic_parser import DotDict
from matplotlib import pyplot as plt, lines as mlines

from omc3.plotting.spectrum.utils import (plot_lines, FigureContainer, get_cycled_color,
                                          get_approx_size_in_axes_coordinates,
                                          PLANES, STEM_LINES_ALPHA, LABEL_Y_SPECTRUM,
                                          LABEL_X, AMPS, FREQS, output_plot)
from omc3.plotting.utils.annotations import get_fontsize_as_float
from omc3.utils import logging_tools

LOG = logging_tools.getLogger(__name__)


def create_stem_plots(figures: dict, opt: DotDict) -> None:
    """ Main loop for stem-plot creation. """
    LOG.debug(f"  ...creating Stem Plots")
    for fig_id, fig_cont in figures.items():
        LOG.debug(f'   Plotting Figure: {fig_id}.')
        fig_cont.fig.canvas.manager.set_window_title(fig_id)

        _plot_stems(fig_cont)
        plot_lines(fig_cont, opt.lines)
        _format_axes(fig_cont, opt.limits)
        if opt.ncol_legend > 0:
            _create_legend(fig_cont.axes[0], fig_cont.data.keys(), opt.lines, opt.ncol_legend)
        output_plot(fig_cont)

    if opt.show:
        plt.show()


def _plot_stems(fig_cont: FigureContainer) -> None:
    """ Plot the spectrum stems for this figure container. """
    for idx_plane, plane in enumerate(PLANES):
        ax = fig_cont.axes[idx_plane]
        for idx_data, (label, data) in enumerate(fig_cont.data.items()):
            if data[plane] is None:
                continue
            # plot
            markers, stems, base = ax.stem(data[plane][FREQS], data[plane][AMPS],
                                           use_line_collection=True, basefmt='none', label=label)

            # Set appropriate colors
            color = get_cycled_color(idx_data)
            markers.set_markeredgecolor(color)
            stems.set_color(color)
            stems.set_alpha(STEM_LINES_ALPHA)
            LOG.debug(f"    {label} {plane}: color={color}, nstems={len(data[plane][FREQS])}")


def _format_axes(fig_cont: FigureContainer, limits: DotDict):
    for idx_plane, plane in enumerate(PLANES):
        ax = fig_cont.axes[idx_plane]

        ax.set_yscale('log')
        ax.set_xlim(limits.xlim)
        ax.set_ylim(limits.ylim)
        ax.set_ylabel(LABEL_Y_SPECTRUM.format(plane=plane.upper()))
        ax.set_xlabel(LABEL_X)
        ax.tick_params(axis='both', which='major')


def _create_legend(ax, labels, lines, ncol):
    lines_params = dict(
        marker=matplotlib.rcParams[u'lines.marker'],
        markersize=matplotlib.rcParams[u'font.size'] * 0.5,
        linestyle='None',
    )
    legend_params = dict(
        loc='lower right',
        ncol=ncol,
        fancybox=False, shadow=False, frameon=False,
    )
    handles = [mlines.Line2D([], [], color=get_cycled_color(idx), label=bpm, **lines_params)
               for idx, bpm in enumerate(labels)]
    leg = ax.legend(handles=handles, **legend_params)

    leg.axes.figure.canvas.draw()  # to get the legend extend

    # check if it is wider than the axes
    legend_width = leg.get_window_extent().transformed(leg.axes.transAxes.inverted()).width
    x_shift = 0
    if legend_width > 1:
        x_shift = (legend_width - 1) / 2.  # shift more into center

    # move above line-labels
    nlines = sum([line is not None for line in lines.values()]) + 0.05
    label_size = get_fontsize_as_float(matplotlib.rcParams['axes.labelsize'])
    y_shift = get_approx_size_in_axes_coordinates(leg.axes, label_size=label_size) * nlines
    leg.axes.legend(handles=handles, bbox_to_anchor=(1. + x_shift, 1. + y_shift), **legend_params)



"""
PS
--

Accelerator-Class for the ``PS`` machine.

Model Creation Keyword Args:
    *--Optional--*

    - **dpp** *(float)*:

        Deltap/p to use.

        default: ``0.0``


    - **driven_excitation** *(str)*:

        Denotes driven excitation by `AC-dipole` (acd) or by `ADT` (adt)

        choices: ``('acd', 'adt')``


    - **drv_tunes** *(float)*:

        Driven tunes without integer part.


    - **energy** *(float)*:

        Energy in **Tev**.


    - **model_dir** *(str)*:

        Path to model directory; loads tunes and excitation from model!


    - **modifiers** *(str)*:

        Path to the optics file to use (modifiers file).


    - **nat_tunes** *(float)*:

        Natural tunes without integer part.


    - **xing**:

        If True, x-ing angles will be applied to model

        action: ``store_true``
"""
import logging
from pathlib import Path

from generic_parser import EntryPoint

from omc3.model.accelerators.accelerator import Accelerator, AccElementTypes, AccExcitationMode
from omc3.model.constants import PLANE_TO_HV, MODIFIER_TAG

LOGGER = logging.getLogger(__name__)
CURRENT_DIR = Path(__file__).parent


class Ps(Accelerator):
    """Parent Class for PS-types."""
    NAME = "ps"
    RE_DICT = {AccElementTypes.BPMS: r"PR\.BPM",
               AccElementTypes.MAGNETS: r".*",
               AccElementTypes.ARC_BPMS: r"PR\.BPM"
               }

    # Public Methods ##########################################################
    @staticmethod
    def get_parameters():
        params = super(Ps, Ps).get_parameters()
        params.add_parameter(
            name="year",
            type=int,
            default=2021,
            choices=(2018, 2021),
            help="Year of the optics.",
        )
        return params

    def __init__(self, *args, **kwargs):
        parser = EntryPoint(self.get_parameters(), strict=True)
        opt = parser.parse(*args, **kwargs)
        super().__init__(opt)
        self.year = opt.year

    def verify_object(self):
        pass

    def get_dir(self):
        return CURRENT_DIR / self.NAME / str(self.year)

    def get_file(self, filename: str) -> Path:
        """Get filepaths for PS files."""
        ps_dir = CURRENT_DIR / self.NAME
        ps_year_dir = self.get_dir()

        for dir_ in (ps_year_dir, ps_dir):
            file_path = dir_ / filename
            if file_path.is_file():
                return file_path

        raise NotImplementedError(
            f"File {file_path.name} not available for accelerator {self.NAME}."
        )

    def get_exciter_bpm(self, plane, bpms):
        if not self.excitation:
            return None
        bpms_to_find = ["PR.BPM00", "PR.BPM03"]
        found_bpms = [bpm for bpm in bpms_to_find if bpm in bpms]
        if not len(found_bpms):
            raise KeyError
        return (list(bpms).index(found_bpms[0]), found_bpms[0]), f"{PLANE_TO_HV[plane]}ACMAP"

    def get_base_madx_script(self, best_knowledge=False):
        if best_knowledge:
            raise NotImplementedError(f"Best knowledge model not implemented for accelerator {self.NAME}")

        use_acd = self.excitation == AccExcitationMode.ACD
        replace_dict = {
            "FILES_DIR": str(self.get_dir()),
            "USE_ACD": str(int(use_acd)),
            "NAT_TUNE_X": self.nat_tunes[0],
            "NAT_TUNE_Y": self.nat_tunes[1],
            "KINETICENERGY": self.energy,
            "DRV_TUNE_X": "",
            "DRV_TUNE_Y": "",
            "MODIFIERS": "",
            "BEAM_FILE": "",  # From 2021
        }
        if self.modifiers:
            replace_dict["MODIFIERS"] = '\n'.join([f" call, file = '{m}'; {MODIFIER_TAG}" for m in self.modifiers])
            replace_dict["BEAM_FILE"] = '\n'.join([f" call, file = '{m.with_suffix('.beam')}';" for m in self.modifiers])
        if use_acd:
            replace_dict["DRV_TUNE_X"] = self.drv_tunes[0]
            replace_dict["DRV_TUNE_Y"] = self.drv_tunes[1]
        mask = self.get_file('base.mask').read_text()
        return mask % replace_dict


class _PsSegmentMixin(object):

    def __init__(self):
        self._start = None
        self._end = None
        self.energy = None



"""
Update Natural Tune in Lin-Files
--------------------------------

Script to update the natural tune in lin files, based on the spectrum data (amps and freqs) and a
given frequency interval.


**Arguments:**

*--Required--*

- **files**: List of paths to the spectrum files.
  The files need to be given without their '.lin'/'.amps[xy]','.freqs[xy]' endings.
  (So usually the path of the TbT-Data file.)

- **interval** *(float)*: Frequency interval in which the highest peak should be found.


*--Optional--*

- **bpms**: List of BPMs which need to be updated. If not given it will be all of them.

- **not_found_action** *(str)*: Defines what to do, if no line was found in given interval.
  'error': throws a ValueError; 'remove': removes the bpm; 'ignore': keeps the old values.

  Choices: ``['error', 'remove', 'ignore']``
  Default: ``error``
- **planes** *(str)*: Which planes.

  Choices: ``('X', 'Y')``
  Default: ``['X', 'Y']``
- **rename_suffix** *(str)*: Additional suffix for output lin-file.
  Will be inserted between filename and extension.
  If empty, the original file is overwritten - unless they are old files,
  then the omc3 filename convention will be used.

  Default: ``None``
"""

from collections import OrderedDict
from contextlib import suppress
from pathlib import Path

import pandas as pd
import tfs
from generic_parser import entrypoint, EntryPointParameters
from generic_parser.entrypoint_parser import save_options_to_config

from omc3.definitions import formats
from omc3.definitions.constants import PLANES
from omc3.harpy.constants import COL_NATTUNE, COL_NATAMP, COL_NAME, FILE_LIN_EXT
from omc3.harpy.handler import _compute_headers
from omc3.plotting.spectrum.utils import (load_spectrum_data, get_bpms,
                                          LIN, AMPS, FREQS
                                          )
from omc3.utils.logging_tools import get_logger, list2str

LOG = get_logger(__name__)


def get_params():
    return EntryPointParameters(
        files=dict(
            required=True,
            nargs='+',
            help=("List of paths to the spectrum files. The files need to be given"
                  " without their '.lin'/'.amps[xy]','.freqs[xy]' endings. "
                  " (So usually the path of the TbT-Data file.)")
        ),
        interval=dict(
            required=True,
            nargs=2,
            type=float,
            help="Frequency interval in which the highest peak should be found."
        ),
        bpms=dict(
            nargs='+',
            help=('List of BPMs which need to be updated. '
                  'If not given it will be all of them.')
        ),
        planes=dict(
            nargs='+',
            type=str,
            help="Which planes.",
            choices=PLANES,
            default=list(PLANES),
        ),
        rename_suffix=dict(
            type=str,
            help=("Additional suffix for output lin-file. "
                  "Will be inserted between filename and extension. "
                  "If empty, the original file is overwritten - unless they "
                  "are old files, then the omc3 filename convention will be "
                  "used."),
            default=''
        ),
        not_found_action=dict(
            type=str,
            choices=['error', 'remove', 'ignore'],
            help=('Defines what to do, if no line was found in given interval.'
                  "'error': throws a ValueError; 'remove': removes the bpm; "
                  "'ignore': keeps the old values."),
            default='error'
        ),
    )


@entrypoint(get_params(), strict=True)
def main(opt):
    _save_options_to_config(opt)
    LOG.info("Updating Natural Tunes in Lin-Files.")

    gathered = [None] * len(opt.files)
    for idx_file, file_path in enumerate(Path(f) for f in opt.files):
        data = load_spectrum_data(file_path, opt.bpms, opt.planes)
        bpms = get_bpms(data[LIN], opt.bpms, file_path, opt.planes)

        data = _update_lin_columns(data, bpms,
                                   opt.planes, opt.interval, opt.not_found_action,
                                   file_path.name)

        data = _update_lin_header(data, opt.planes, file_path.name)

        _save_linfiles(data[LIN], file_path, opt.planes, opt.rename_suffix)
        gathered[idx_file] = data[LIN]
    return gathered


# Update -----------------------------------------------------------------------


def _update_lin_columns(data, bpms, planes, interval, not_found_action, filename):
    for plane in planes:
        col_nattune = f'{COL_NATTUNE}{plane.upper()}'
        col_natamp = f'{COL_NATAMP}{plane.upper()}'

        for bpm in bpms[plane]:
            freqs, amps = data[FREQS][plane][bpm], data[AMPS][plane][bpm]
            peak = _get_peak_in_interval(freqs, amps, interval)

            if peak is None:
                msg = (f'No lines found for bpm {bpm} in plane {plane} '
                       f'in interval {list2str(interval)} for file-id "{filename}".')
                if not_found_action == 'error':
                    raise ValueError(msg)
                LOG.warning(msg)

                if not_found_action == 'remove':
                    data[LIN][plane] = data[LIN][plane].drop(bpm, axis='index')

            else:
                freq_peak, amp_peak = peak
                data[LIN][plane].loc[bpm, col_nattune] = freq_peak
                data[LIN][plane].loc[bpm, col_natamp] = amp_peak
                LOG.debug(f"{filename}.{bpm}.{plane} nattune set to"
                          f"f={freq_peak}, A={amp_peak}")
    return data


def _get_peak_in_interval(freqs, amps, interval):
    data_series = pd.Series(data=amps.to_numpy(), index=freqs.to_numpy())
    data_series = data_series.sort_index()
    try:
        f_peak = data_series.loc[slice(*sorted(interval))].idxmax()
    except ValueError:
        return None
    else:
        return f_peak, data_series.loc[f_peak]


def _update_lin_header(data, planes, filename):
    for plane in planes:
        LOG.debug(f"Updating headers for {filename}, plane {plane}")
        data[LIN][plane].headers.update(_compute_headers(data[LIN][plane]))
    return data


# Output -----------------------------------------------------------------------


def _save_linfiles(lin_data, file_path, planes, id_suffix):
    for plane in planes:
        out_path = file_path.with_name(f'{file_path.name}{id_suffix}{FILE_LIN_EXT.format(plane=plane.lower())}')
        tfs.write(str(out_path), lin_data[plane], save_index=COL_NAME)


def _save_options_to_config(opt):
    with suppress(IOError):
        save_options_to_config(formats.get_config_filename(__file__),
                               OrderedDict(sorted(opt.items())))


# Script Mode ------------------------------------------------------------------


if __name__ == '__main__':
    main()



"""
Response Matrix IO
------------------

Input and output functions for response matrices.
"""
from collections import defaultdict
from pathlib import Path
from typing import Dict, Sequence, Set

import pandas as pd
import logging


LOG = logging.getLogger(__name__)

COMPLIB = 'blosc'  # zlib is the standard compression
COMPLEVEL = 9  # goes from 0-9, 9 is highest compression, None deactivates if COMPLIB is None


# Fullresponse -----------------------------------------------------------------


def read_fullresponse(path: Path, optics_parameters: Sequence[str] = None) -> Dict[str, pd.DataFrame]:
    """Load the response matrices from disk.
    Beware: As empty DataFrames are skipped on write,
    default for not found entries are empty DataFrames.
    """
    if not path.exists():
        raise IOError(f"Fullresponse file {str(path)} does not exist.")

    LOG.info(f"Loading response matrices from file '{str(path)}'")
    with pd.HDFStore(path, mode='r') as store:
        _check_keys(store, optics_parameters, 'fullresponse')

        fullresponse = defaultdict(pd.DataFrame)
        if optics_parameters is None:
            optics_parameters = _main_store_groups(store)
        for p in optics_parameters:
            fullresponse[p] = store[p]
    return fullresponse


def write_fullresponse(path: Path, fullresponse: Dict[str, pd.DataFrame]):
    """Write the full response matrices to disk.
    Beware: Empty Dataframes are skipped! (HDF creates gigantic files otherwise)
    """
    LOG.info(f"Saving response matrices into file '{str(path)}'")
    if path.exists():
        LOG.warning(f"Fullresponse file {str(path)} already exist and will be overwritten.")
    with pd.HDFStore(path, mode='w', complib=COMPLIB, complevel=COMPLEVEL) as store:
        for param, response_df in fullresponse.items():
            store.put(value=response_df, key=param, format="table")


# Varmap -----------------------------------------------------------------------


def read_varmap(path: Path, k_values: Sequence[str] = None) -> Dict[str, Dict[str, pd.Series]]:
    """Load the variable mapping file from disk.
    Beware: As empty DataFrames are skipped on write,
    default for not found entries are empty Series.
    """
    if not path.exists():
        raise IOError(f"Varmap file {str(path)} does not exist.")

    LOG.info(f"Loading varmap from file '{str(path)}'")
    with pd.HDFStore(path, mode='r') as store:
        _check_keys(store, k_values, 'varmap')

        varmap = defaultdict(lambda: defaultdict(pd.Series))
        for key in store.keys():
            _, param, subparam = key.split('/')
            if k_values is not None and param not in k_values:
                continue
            varmap[param][subparam] = store[key]
    return varmap


def write_varmap(path: Path, varmap: Dict[str, Dict[str, pd.Series]]):
    """Write the  variable mapping file to disk.
    Beware: Empty Dataframes are skipped! (HDF creates gigantic files otherwise)
    """
    LOG.info(f"Saving varmap into file '{str(path)}'")
    with pd.HDFStore(path, mode='w', complib=COMPLIB, complevel=COMPLEVEL) as store:
        for param, sub in varmap.items():
            for subparam, varmap_series in sub.items():
                store.put(value=varmap_series, key=f"{param}/{subparam}", format="table")


# Helper -----------------------------------------------------------------------


def _check_keys(store: pd.HDFStore, keys: Sequence[str], id:str):
    if keys is None:
        return

    groups = _main_store_groups(store)
    not_found = [k for k in keys if k not in groups]
    if len(not_found):
        raise ValueError(f"The following keys could not be found in {id} file:"
                         f" {', '.join(not_found)}")


def _main_store_groups(store: pd.HDFStore) -> Set[str]:
    """Returns sequence of unique main store groups."""
    return {k.split('/')[1] for k in store.keys()}



"""
Contexts
--------

Provides contexts managers to use.
"""
import os
import shutil
import sys
import tempfile
import time
import warnings
from contextlib import contextmanager


@contextmanager
def log_out(stdout=sys.stdout, stderr=sys.stderr):
    """Temporarily changes ``sys.stdout`` and ``sys.stderr``."""
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    sys.stdout = stdout
    sys.stderr = stderr
    try:
        yield
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr


@contextmanager
def silence():
    """
    Suppress all console output, rerouting ``sys.stdout`` and ``sys.stderr`` to devnull.
    """
    devnull = open(os.devnull, "w")
    with log_out(stdout=devnull, stderr=devnull):
        try:
            yield
        finally:
            devnull.close()


@contextmanager
def timeit(function):
    """Prints the time at the end of the context via ``function``."""
    start_time = time.time()
    try:
        yield
    finally:
        time_used = time.time() - start_time
        function(time_used)


@contextmanager
def suppress_warnings(warning_classes):
    """Suppress all warnings of given classes."""
    with warnings.catch_warnings(record=True) as warn_list:
        yield
    for w in warn_list:
        if not issubclass(w.category, warning_classes):
            print("{file:s}:{line:d}: {clas:s}: {message:s}".format(
                file=w.filename,
                line=w.lineno,
                clas=w._category_name,
                message=w.message.message,
            ),
                file=sys.stderr
            )



"""
Plotting Utilities: Colors
--------------------------

Helper functions to handle colors in plots.
"""
import colorsys
from itertools import cycle

import matplotlib
from matplotlib import colors as mc


def get_mpl_color(idx=None):
    """Gets the 'new' default ``matplotlib`` colors by index, or the whole cycle."""
    c = [
        '#1f77b4',  # muted blue
        '#ff7f0e',  # safety orange
        '#2ca02c',  # cooked asparagus green
        '#d62728',  # brick red
        '#9467bd',  # muted purple
        '#8c564b',  # chestnut brown
        '#e377c2',  # raspberry yogurt pink
        '#7f7f7f',  # middle gray
        '#bcbd22',  # curry yellow-green
        '#17becf',  # blue-teal
    ]
    if idx is None:
        return cycle(c)
    return c[idx % len(c)]


def rgb_plotly_to_mpl(rgb_string):
    """Helper function to transforn plotly rbg to ``matplotlib`` rgb."""
    if rgb_string.startswith('#'):
        return rgb_string

    rgb_string = rgb_string.replace("rgba", "").replace("rgb", "")
    rgb = eval(rgb_string)
    rgb_norm = [c/255. for c in rgb]
    return rgb_norm


def change_color_brightness(color, amount=0.5):
    """
    Lightens the given color by multiplying (1-luminosity) by the given amount. Input can be
    ``matplotlib`` color string, hex string, or RGB tuple. An amount of 1 equals to no change. 0
    is very bright (white) and 2 is very dark. Original code by Ian Hincks
    Source: https://stackoverflow.com/questions/37765197/darken-or-lighten-a-color-in-matplotlib
    """
    if not (0<=amount<=2):
        raise ValueError("The brightness change has to be between 0 and 2."
                         " Instead it was {}".format(amount))
    try:
        c = mc.cnames[color]
    except KeyError:
        c = color

    try:
        c = colorsys.rgb_to_hls(*mc.ColorConverter().to_rgb(c))  # matplotlib 1.5
    except AttributeError:
        c = colorsys.rgb_to_hls(*mc.to_rgb(c))  # matplotlib > 2
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def change_ebar_alpha_for_line(ebar, alpha):
    """
    Changes the alpha value of an error-bar container.
    Loop through bars (ebar[1]) and caps (ebar[2]) and set the alpha value.
    """
    for bars_or_caps in ebar[1:]:
        for bar_or_cap in bars_or_caps:
            bar_or_cap.set_alpha(alpha)


def change_ebar_alpha_for_axes(ax, alpha):
    """Wrapper for change_ebar_alpha_for_line to change all in one axes."""
    for ebar in ax.containers:
        if isinstance(ebar, matplotlib.container.ErrorbarContainer):
            change_ebar_alpha_for_line(ebar, alpha)


"""
Constants
---------

This module provides high-level functions to manage most functionality of ``model``.
Specific constants to be used in ``model``, to help with consistency.
"""

from pathlib import Path

MACROS_DIR = "macros"
OBS_POINTS = "observation_points.def"
MODIFIERS_MADX = 'modifiers.madx'
MODIFIER_TAG = "!@modifier"
TWISS_BEST_KNOWLEDGE_DAT = "twiss_best_knowledge.dat"
TWISS_ELEMENTS_BEST_KNOWLEDGE_DAT = "twiss_elements_best_knowledge.dat"
TWISS_ADT_DAT = "twiss_adt.dat"
TWISS_AC_DAT = "twiss_ac.dat"
TWISS_ELEMENTS_DAT = "twiss_elements.dat"
TWISS_DAT = "twiss.dat"
ERROR_DEFFS_TXT = "error_deffs.txt"
JOB_MODEL_MADX = "job.create_model.madx"

GENERAL_MACROS = "general.macros.madx"
LHC_MACROS = "lhc.macros.madx"
LHC_MACROS_RUN3 = "lhc.macros.run3.madx"

B2_SETTINGS_MADX = "b2_settings.madx"
B2_ERRORS_TFS = "b2_errors.tfs"
PLANE_TO_HV = dict(X="H", Y="V")

ACCELERATOR_MODEL_REPOSITORY = Path("/afs/cern.ch/eng/acc-models/lhc")


"""
Linfile Cleaning
----------------
Performs an automated cleaning of different columns in the lin-file
as a standalone script to allow for manual refinement after harpy is done.

All data is assumed to be gaussian distributed around a "true" value,
and outliers are cleaned by calculating average and standard deviation
of the given data.

The cleaning is done by removing all data-points that are outside of the
1-0.5/n estimated percentile of the data. Where n is the number of
(remaining) data-points in each loop, and the process is repeated until
n stays constant (or 2 or less data-points remain).

Datapoints with a standard deviation smaller than the given limit are not
cleaned. The limit is given in whatever units the data itself is in and
is an absolute value.

Cleaning is done per given file independently
i.e. removed BPMs from one file can be present in the next.
The columns are iterated on the same file, i.e. the cleaning is done
consecutively on the already cleaned data from the last column,
yet the moments of the distribution themselves are evaluated per column.

In the end, the lin-files are overwritten with the cleaned ones.
If the ``backup`` option is activated, a backup of the original file
is created, which can be restored via the ``restore`` option.
In case the restore-flag is given, only the filenames are required.
No cleaning is performed with this option.

**Arguments:**

*--Required--*

- **files** *(PathOrStr)*:

    List of paths to the lin-files, including suffix.


*--Optional--*

- **backup**:

    Create a backup of the files before overwriting. The backups are
    numbered, with the highest number being the latest backup.

    action: ``store_true``


- **columns** *(str)*:

    Columns to clean on.


- **limit** *(float)*:

    Do not clean data-points deviating less than this limit from the
    average.

    default: ``0.0``


- **restore**:

    Restore the latest backup of the files. If this parameter is given, no
    cleaning is performed.

    action: ``store_true``


TODO:
also use isolation forest, BUT this probably needs some modification there
as well, as it only cleans on TUNE, not on NATTUNE.
And it requires an accelerator instance.
"""
import shutil
from numbers import Number
from pathlib import Path
from typing import Sequence, Union

import pandas as pd
import tfs
from generic_parser.entrypoint_parser import (
    entrypoint, EntryPointParameters
)

from omc3.definitions.formats import BACKUP_FILENAME
from omc3.harpy.constants import COL_NAME
from omc3.harpy.handler import _compute_headers
from omc3.utils import logging_tools
from omc3.utils.iotools import PathOrStr, save_config
from omc3.utils.outliers import get_filter_mask

LOG = logging_tools.get_logger(__name__)


def get_params():
    return EntryPointParameters(
        files=dict(
            required=True,
            type=PathOrStr,
            nargs='+',
            help="List of paths to the lin-files, including suffix.",
        ),
        # restore backup
        restore=dict(
            action="store_true",
            help=("Restore the latest backup of the files. "
                  "If this parameter is given, no cleaning is performed."),
        ),
        # for actual cleaning
        columns=dict(
            nargs='+',
            type=str,
            help="Columns to clean on.",
        ),
        limit=dict(
            type=float,
            help="Do not clean data-points deviating less than this limit from the average.",
            default=0.0,
        ),
        backup=dict(
            action="store_true",
            help=("Create a backup of the files before overwriting. "
                  "The backups are numbered, with the highest number being "
                  "the latest backup.")
        ),
    )


@entrypoint(get_params(), strict=True)
def main(opt):
    """Main function, to parse commandline input and separate restoration
    from cleaning."""
    save_config(Path('.'), opt, __file__)
    if opt.restore:
        restore_files(opt.files)
        return

    if opt.columns is None:
        raise ValueError("The option 'columns' is required for cleaning.")
    clean_columns(opt.files, opt.columns, opt.limit, opt.backup)


# Restore ----------------------------------------------------------------------

def restore_files(files: Sequence[Union[Path, str]]):
    """Restore backupped files."""
    failed = []
    for file in files:
        file = Path(file)
        try:
            _restore_file(file)
        except IOError as e:
            failed.append(str(e))

    if len(failed):
        all_errors = '\n'.join(failed)
        if len(failed) == len(files):
            raise IOError(f"Restoration of ALL files has failed\n{all_errors}")
        raise IOError(f"Restoration of some files has failed, "
                      f"but the others were restored:\n{all_errors}")

    LOG.info("Restoration successfully completed.")


def _restore_file(file):
    counter = 1
    backup_file = _get_backup_filepath(file, counter)
    if not backup_file.exists():
        raise IOError(f"No backups found for file {file.name}")

    # get last existing backup file
    while backup_file.exists():
        counter += 1
        backup_file = _get_backup_filepath(file, counter)
    backup_file = _get_backup_filepath(file, counter-1)

    # restore found  file
    LOG.info(f"Restoring last backup file {backup_file.name} to {file.name}.")
    shutil.move(backup_file, file)  # overwrites file


# Clean ------------------------------------------------------------------------

def clean_columns(files: Sequence[Union[Path, str]], columns: Sequence[str],
                  limit: float = 0.0, backup: bool = True):
    """ Clean the columns in the given files."""
    for file in files:
        file = Path(file)
        LOG.info(f"Cleaning {file.name}.")
        df = tfs.read_tfs(file, index=COL_NAME)
        for column in columns:
            df = _filter_by_column(df, column, limit)
        df.headers.update(_compute_headers(df))

        if backup:
            _backup_file(file)

        tfs.write_tfs(file, df, save_index=COL_NAME)


def _filter_by_column(df: pd.DataFrame, column: str, limit: Number) -> pd.DataFrame:
    """Get the dataframe with all rows dropped filtered by the given column."""
    if column not in df.columns:
        LOG.info(f"{column} not in current file. Skipping cleaning.")
        return df

    good_bpms = get_filter_mask(data=df[column], limit=limit)
    n_good, n_total = sum(good_bpms), len(good_bpms)
    LOG.info(f"Cleaned {n_total-n_good:d} of {n_total:d} elements in {column} ({n_good:d} remaining).")
    return df.loc[good_bpms, :]


# Backup ---

def _backup_file(file):
    counter = 1
    backup_file = _get_backup_filepath(file, counter)
    while backup_file.exists():
        counter += 1
        backup_file = _get_backup_filepath(file, counter)

    LOG.info(f"Backing up original file {file.name} to {backup_file.name}.")
    shutil.copy(file, backup_file)


def _get_backup_filepath(file: Path, counter: int):
    return file.with_name(BACKUP_FILENAME.format(basefile=file.name, counter=counter))


# Script Mode ------------------------------------------------------------------

if __name__ == '__main__':
    main()



r"""
Response TWISS
--------------

Provides a class to get response matrices from Twiss parameters.

.. warning::
  The responses are only valid for MAD-X Beam 1 and Beam 2 twiss-files, **not for Beam 4** !!
  Also, it only works properly for on-orbit twiss files.


The calculation is based on formulas in [#FranchiAnalyticformulasrapid2017]_, [#TomasReviewlinearoptics2017]_
and is summarized in [#DillyUpdatedGlobalOpticsCorrection2018]_, where the following
equations can be found in Eq. 10 - Eq. 15.


* Beta Response:

.. math::

    \delta \beta_{z,j} = \mp \beta_{z,j} \sum_m \delta K_{1,m} \frac{\beta_{z,m}}{2}
    \frac{cos(2\tau_{z,mj})}{sin(2\pi Q_z)}


* Dispersion Response:

.. math::

    \delta D_{x,j} =&+ \sqrt{\beta_{x,j}} \sum_m (\delta K_{0,m} + \delta K_{1S,m} D_{y,m}
    - \delta K_{1,m} D_{x,m}) \frac{\sqrt{\beta_{x,m}}}{2}
    \frac{cos(\tau_{x,mj})}{sin(\pi Q_x)}
    \\
    \delta D_{y,j} =&- \sqrt{\beta_{y,j}} \sum_m (\delta K_{0S,m}
    - \delta K_{1S,m} D_{x,m} - \delta K_{1,m} D_{y,m}) \frac{\sqrt{\beta_{y,m}}}{2}
    \frac{cos(\tau_{y,mj})}{sin(\pi Q_y)}


* Norm. Dispersion Response: similar as above but with :math:`\frac{1}{\sqrt{\beta}}` linearized

.. math::

    \delta \frac{D_{x,j}}{\sqrt{\beta_{x,j}}} =&+ \sum_m (\delta K_{0,m} + \delta K_{1S,m} D_{y,m}
    - \delta K_{1,m} D_{x,m} ) \frac{\sqrt{\beta_{x,m}}}{2}
    \frac{cos(\tau_{x,mj})}{sin(\pi Q_x)}
    &&+ \frac{D_{x,j}}{\sqrt{\beta_{x,j}}} \delta K_{1,m}
    \frac{\beta_{x,m}}{4}\frac{cos(2\tau_{x,mj})}{2sin(\pi Q_x)}
    \\
    \delta \frac{D_{y,j}}{\sqrt{\beta_{y,j}}} =&- \sum_m (\delta K_{0S,m} - \delta K_{1S,m} D_{x,m}
    - \delta K_{1,m} D_{y,m}) \frac{\sqrt{\beta_{y,m}}}{2}
    \frac{cos(\tau_{y,mj})}{sin(\pi Q_y)}
    &&- \frac{D_{y,j}}{\sqrt{\beta_{y,j}}} \delta K_{1,m}
    \frac{\beta_{y,m}}{4}\frac{cos(2\tau_{y,mj})}{2sin(\pi Q_y)}


* Phase Advance Response:

.. math::

    \delta \Phi_{z,wj} = \pm \sum_m \delta K_{1,m} \frac{\beta_{z,m}}{4}
    \left\{ 2\left[ \Pi_{mj} - \Pi_{mw} + \Pi_{jw} \right] +
    \frac{sin(2\tau_{z,mj}) - sin(2\tau_{z,mw})}{sin(2\pi Q_z)} \right\}


* Tune Response:

.. math::

    \delta Q_z = \pm \sum_m \delta K_{1,m} \frac{\beta_{z,m}}{4\pi}


* Coupling Response:

.. math::

    \delta f_{\substack{\scriptscriptstyle 1001 \\ \scriptscriptstyle 1010},j} =
    \sum_m \delta J_{1,m} \, \frac{\sqrt{\beta_{x,m}\beta_{y,m}}}{4} \,
    \frac{\exp{(i(\Delta\Phi_{x,mj} \mp \Delta\Phi_{y,mj}))}}{1-\exp({2\pi i (Q_x \mp Q_y}))}

For people reading the code, the response matrices are first calculated like:

::

    |  Elements of interest (j) --> ... |
    |Magnets (m)                        |
    |  |                                |
    |  v                                |
    |  .                                |
    |  .                                |
    |  .                                |
    |                                   |

This avoids transposing all vectors individually in the beginning.
At the end (of the calculation) the matrix is then transposed
to fit the :math:`M \cdot \delta K` orientation.

Also :math:`\Delta \Phi_{z,wj}` needs to be multiplied by :math:`2\pi` to be consistent.


.. rubric:: References

..  [#FranchiAnalyticformulasrapid2017]
    A. Franchi et al.,
    Analytic formulas for the rapid evaluation of the orbit response matrix
    and chromatic functions from lattice parameters in circular accelerators. (2017)
    https://arxiv.org/abs/1711.06589

.. [#TomasReviewlinearoptics2017]
    R. Tomas, et al.,
    'Review of linear optics measurement and correction for charged particle
    accelerators.'
    Physical Review Accelerators and Beams, 20(5), 54801. (2017)
    https://doi.org/10.1103/PhysRevAccelBeams.20.054801

..  [#DillyUpdatedGlobalOpticsCorrection2018]
    J. Dilly et al.,
    An updated global optics correction scheme. (2018)
    https://cds.cern.ch/record/2632945/

"""
import copy
from typing import Dict, List, Sequence

import numpy as np
import pandas as pd
import tfs

from omc3.correction.constants import BETA, DISP, F1001, F1010, NORM_DISP, PHASE_ADV, TUNE, S, PHASE
from omc3.correction.response_io import read_varmap
from omc3.correction.sequence_evaluation import check_varmap_file
from omc3.model.accelerators.accelerator import Accelerator
from omc3.utils import logging_tools
from omc3.utils.contexts import timeit

LOG = logging_tools.get_logger(__name__)

DUMMY_ID = "DUMMY_PLACEHOLDER"
PLANES = ("X", "Y")

# Twiss Response Class ########################################################


class TwissResponse:
    """Provides Response Matrices calculated from sequence, model and given variables.

    Args:
        accel_inst (accelerator): Accelerator Instance (needs to contain elements model).
        variable_categories (list): List of variable categories to get from the accelerator class.
        varmap_or_path (dict, string): mapping of the variables,
            either as dict-structure of Series or path to a pickled-file.
        at_elements (str): Get response matrix for these elements. Can be:
            'bpms': All BPMS (Default)
            'bpms+': BPMS+ used magnets (== magnets defined by variables in varfile)
            'all': All BPMS and Magnets given in the model (Markers are removed)

    """

    ################################
    #            INIT
    ################################

    def __init__(self, accel_inst, variable_categories, varmap_or_path, at_elements="bpms"):

        LOG.debug("Initializing TwissResponse.")
        with timeit(lambda t: LOG.debug(f"  Time initializing TwissResponse: {t} s")):
            # Get input
            self._twiss = self._get_model_twiss(accel_inst)
            self._beam_sign = 1 if getattr(accel_inst, "beam", 1) == 1 else -1
            self._variables = accel_inst.get_variables(classes=variable_categories)
            self._var_to_el = self._get_variable_mapping(varmap_or_path)
            self._elements_in = self._get_input_elements()
            self._elements_out = self._get_output_elements(at_elements)

            # calculate all phase advances
            self._phase_advances = get_phase_advances(self._twiss)

            # All responses are calcluated as needed, see getters below!
            # slots for response matrices
            self._beta = None
            self._dispersion = None
            self._phase = None
            self._phase_adv = None
            self._tune = None
            self._coupling = None
            self._beta_beat = None
            self._norm_dispersion = None

            # slots for mapped response matrices
            self._coupling_mapped = None
            self._beta_mapped = None
            self._dispersion_mapped = None
            self._phase_mapped = None
            self._phase_adv_mapped = None
            self._tune_mapped = None
            self._beta_beat_mapped = None
            self._norm_dispersion_mapped = None

    @staticmethod
    def _get_model_twiss(accel_inst):
        """ Load model, but keep only BPMs and Magnets """
        model = accel_inst.elements
        LOG.debug("Removing non-necessary entries:")
        LOG.debug(f"  Entries total: {model.shape[0]}")
        mask = accel_inst.get_element_types_mask(model.index, types=["bpm", "magnet"])
        model = model.loc[mask, :].copy()  # make a copy to suppress "SettingWithCopyWarning"
        LOG.debug(f"  Entries left: {model.shape[0]}")

        # Add Dummy for Phase Calculations
        model.loc[DUMMY_ID, [f"{S}", f"{PHASE_ADV}X", f"{PHASE_ADV}Y"]] = 0.0
        return model

    def _get_variable_mapping(self, varmap_or_path):
        """Get variable mapping as dictionary

        Dev hint: Define _variables first!
        """
        LOG.debug("Converting variables to magnet names.")
        variables = self._variables

        if not len(variables):
            raise ValueError("No variables found. Maybe wrong categories?")

        try:
            mapping = read_varmap(varmap_or_path)
        except AttributeError:
            LOG.debug("Received varmap as dictionary.")
            mapping = varmap_or_path
        else:
            LOG.debug(f"Loaded varmap from file '{varmap_or_path}'")

        # check if all variables can be found
        check_var = [var for var in variables if all(var not in mapping[order] for order in mapping)]
        if len(check_var) > 0:
            raise ValueError(f"Variables '{', '.join(check_var)}' cannot be found in sequence!")

        # drop mapping for unused variables
        [mapping[order].pop(var) for order in mapping for var in mapping[order].keys() if var not in variables]
        return mapping

    def _get_input_elements(self):
        """Return variable names of input elements.

        Dev hint: Define _var_to_el and _twiss first!
        """
        v2e = self._var_to_el
        tw = self._twiss

        el_in = dict.fromkeys(k for k in v2e.keys() if k[-1] == 'L')
        for order in el_in:
            el_order = []
            for var in v2e[order]:
                el_order += upper(v2e[order][var].index)
            el_in[order] = tw.loc[list(set(el_order)), f"{S}"].sort_values().index.tolist()

        for order in ["K0L", "K0SL", "K1L", "K1SL"]:
            if order not in el_in:
                el_in[order] = []
        return el_in

    def _get_output_elements(self, at_elements):
        """Return name-array of elements to use for output.

        Dev hint: Define _elements_in first!
        """
        tw_idx = self._twiss.index

        if isinstance(at_elements, list):
            # elements specified
            if any(el not in tw_idx for el in at_elements):
                LOG.warning("One or more specified elements are not in the model.")
            return [idx for idx in tw_idx if idx in at_elements]

        if at_elements == "bpms":
            # bpms only
            return [idx for idx in tw_idx if idx.upper().startswith("B")]

        if at_elements == "bpms+":
            # bpms and the used magnets
            el_in = self._elements_in
            return [idx for idx in tw_idx
                    if (idx.upper().startswith('B') or any(idx in el_in[order] for order in el_in))]

        if at_elements == "all":
            # all, obviously
            return [idx for idx in tw_idx if idx != DUMMY_ID]

    ################################
    #       Response Matrix
    ################################

    def _calc_coupling_response(self):
        """Response Matrix for coupling."""
        LOG.debug("Calculate Coupling Matrix")
        with timeit(lambda t: LOG.debug(f"  Time needed: {t} s")):
            tw = self._twiss
            adv = self._phase_advances
            el_out = self._elements_out
            k1s_el = self._elements_in["K1SL"]
            dcoupl = dict.fromkeys(["1001", "1010"])

            i2pi = 2j * np.pi
            phx = dphi(adv["X"].loc[k1s_el, el_out], tw.Q1).to_numpy()
            phy = dphi(adv["Y"].loc[k1s_el, el_out], tw.Q2).to_numpy()
            bet_term = np.sqrt(tw.loc[k1s_el, f"{BETA}X"].to_numpy() * tw.loc[k1s_el, f"{BETA}Y"].to_numpy())

            for coupling_rdt in ["1001", "1010"]:
                phs_sign = -1 if coupling_rdt == "1001" else 1
                dcoupl[coupling_rdt] = pd.DataFrame(
                    bet_term[:, None] * np.exp(i2pi * (phx + phs_sign * phy)) /
                    (4 * (1 - np.exp(i2pi * (tw.Q1 + phs_sign * tw.Q2)))),
                    index=k1s_el, columns=el_out).transpose()
        return dcoupl

    def _calc_beta_response(self):
        """Response Matrix for delta beta."""
        LOG.debug("Calculate Beta Response Matrix")
        with timeit(lambda t: LOG.debug(f"  Time needed: {t} s")):
            tw = self._twiss
            adv = self._phase_advances
            el_out = self._elements_out
            k1_el = self._elements_in["K1L"]
            dbeta = dict.fromkeys(PLANES)

            for plane in PLANES:
                col_beta = f"{BETA}{plane}"
                q = tw.Q1 if plane == "X" else tw.Q2
                coeff_sign = -1 if plane == "X" else 1

                pi2tau = 2 * np.pi * tau(adv[plane].loc[k1_el, el_out], q)

                dbeta[plane] = pd.DataFrame(
                    tw.loc[el_out, col_beta].to_numpy()[None, :]
                    * tw.loc[k1_el, col_beta].to_numpy()[:, None]
                    * np.cos(2 * pi2tau.to_numpy())
                    * (coeff_sign / (2 * np.sin(2 * np.pi * q))),
                    index=k1_el,
                    columns=el_out,
                ).transpose()

        return dbeta

    def _calc_dispersion_response(self):
        """Response Matrix for delta normalized dispersion."""
        LOG.debug("Calculate Dispersion Response Matrix")
        with timeit(lambda t: LOG.debug(f"  Time needed: {t} s")):
            tw = self._twiss
            adv = self._phase_advances
            el_out = self._elements_out
            els_in = self._elements_in

            sign_map = {
                "X": {"K0L": 1, "K1L": -1, "K1SL": 1},
                "Y": {"K0SL": -1, "K1L": 1, "K1SL": 1},
            }
            col_disp_map = {
                "X": {"K1L": f"{DISP}X", "K1SL": f"{DISP}Y"},
                "Y": {"K1L": f"{DISP}Y", "K1SL": f"{DISP}X"},
            }
            q_map = {"X": tw.Q1, "Y": tw.Q2}
            disp_resp = dict.fromkeys([f"{p}_{t}" for p in sign_map for t in sign_map[p]])

            for plane in sign_map:
                q = q_map[plane]
                col_beta = f"{BETA}{plane}"
                el_types = sign_map[plane].keys()
                els_per_type = [els_in[el_type] for el_type in el_types]

                coeff = np.sqrt(tw.loc[el_out, col_beta].to_numpy()) / (2 * np.sin(np.pi * q))

                for el_in, el_type in zip(els_per_type, el_types):
                    coeff_sign = sign_map[plane][el_type]
                    out_str = f"{plane}_{el_type}"

                    if len(el_in):
                        pi2tau = 2 * np.pi * tau(adv[plane].loc[el_in, el_out], q)
                        bet_term = np.sqrt(tw.loc[el_in, col_beta])

                        try:
                            col_disp = col_disp_map[plane][el_type]
                        except KeyError:
                            pass
                        else:
                            bet_term *= tw.loc[el_in, col_disp]

                        disp_resp[out_str] = (
                            coeff_sign * coeff[None, :] * bet_term.to_numpy()[:, None] * np.cos(pi2tau)
                        ).transpose()
                    else:
                        LOG.debug(f"  No '{el_type}' variables found. "
                                  f"Dispersion Response '{out_str}' will be empty.")
                        disp_resp[out_str] = pd.DataFrame(None, index=el_out)
        return disp_resp

    def _calc_norm_dispersion_response(self):
        """Response Matrix for delta normalized dispersion."""
        LOG.debug("Calculate Normalized Dispersion Response Matrix")
        with timeit(lambda t: LOG.debug(f"  Time needed: {t} s")):
            tw = self._twiss
            adv = self._phase_advances
            el_out = self._elements_out
            els_in = self._elements_in

            sign_map = {
                "X": {"K0L": 1, "K1L": -1, "K1SL": 1},
                "Y": {"K0SL": -1, "K1L": 1, "K1SL": 1},
            }
            col_disp_map = {
                "X": {"K1L": f"{DISP}X", "K1SL": f"{DISP}Y"},
                "Y": {"K1L": f"{DISP}Y", "K1SL": f"{DISP}X"},
            }
            sign_correct_term = {
                "X": {"K1L": 1},
                "Y": {"K1L": -1},
            }
            q_map = {"X": tw.Q1, "Y": tw.Q2}
            disp_resp = dict.fromkeys(["{p:s}_{t:s}".format(p=p, t=t) for p in sign_map for t in sign_map[p]])

            for plane in sign_map:
                q = q_map[plane]
                col_beta = f"{BETA}{plane}"
                el_types = sign_map[plane].keys()
                els_per_type = [els_in[el_type] for el_type in el_types]

                coeff = 1 / (2 * np.sin(np.pi * q))
                coeff_corr = 1 / (4 * np.sin(2 * np.pi * q))
                for el_in, el_type in zip(els_per_type, el_types):
                    coeff_sign = sign_map[plane][el_type]
                    out_str = "{p:s}_{t:s}".format(p=plane, t=el_type)

                    if len(el_in):
                        pi2tau = 2 * np.pi * tau(adv[plane].loc[el_in, el_out], q)
                        beta_in = tw.loc[el_in, col_beta]
                        bet_term = np.sqrt(beta_in)

                        try:
                            col_disp = col_disp_map[plane][el_type]
                        except KeyError:
                            pass
                        else:
                            bet_term *= tw.loc[el_in, col_disp]

                        result = (coeff_sign * coeff * bet_term).to_numpy()[:, None] * np.cos(pi2tau)

                        # correction term
                        try:
                            sign_corr = sign_correct_term[plane][el_type]
                        except KeyError:
                            pass
                        else:
                            norm_disp_corr = (tw.loc[el_out, col_disp] /
                                          np.sqrt(tw.loc[el_out, col_beta]))
                            result += (sign_corr * coeff_corr * norm_disp_corr.to_numpy()[None, :] *
                                       beta_in.to_numpy()[:, None] * np.cos(2 * pi2tau))

                        disp_resp[out_str] = result.transpose()
                    else:
                        LOG.debug(
                            f"  No '{el_type:s}' variables found. "
                            f"Normalized Dispersion Response '{out_str:s}' will be empty."
                        )
                        disp_resp[out_str] = pd.DataFrame(None, index=el_out)
        return disp_resp

    def _calc_phase_advance_response(self):
        """Response Matrix for delta DPhi.

        Reduced to only phase advances between consecutive elements,
        as the 3D-Matrix of all elements exceeds memory space
        (~11000^3 = 1331 Giga Elements)
        --> w = j-1:  DPhi(z,j) = DPhi(x, (j-1)->j)
        """
        LOG.debug("Calculate Phase Advance Response Matrix")
        with timeit(lambda elapsed: LOG.debug(f"  Time needed: {elapsed} s")):
            tw = self._twiss
            adv = self._phase_advances
            k1_el = self._elements_in["K1L"]

            el_out_all = [DUMMY_ID] + self._elements_out  # Add MU[XY] = 0.0 to the start
            el_out = el_out_all[1:]  # in these we are actually interested
            el_out_mm = el_out_all[0:-1]  # elements--

            if len(k1_el) > 0:
                dmu = dict.fromkeys(PLANES)

                pi = pd.DataFrame(tw[f"{S}"].to_numpy()[:, None] < tw[f"{S}"].to_numpy()[None, :],  # pi(i,j) = s(i) < s(j)
                                      index=tw.index, columns=tw.index, dtype=int)

                pi_term = (
                    pi.loc[k1_el, el_out].to_numpy()
                    - pi.loc[k1_el, el_out_mm].to_numpy()
                    + np.diag(pi.loc[el_out, el_out_mm].to_numpy())[None, :]
                )

                for plane in PLANES:
                    col_beta = f"{BETA}{plane}"
                    q = tw.Q1 if plane == "X" else tw.Q2
                    coeff_sign = 1 if plane == "X" else -1

                    pi2tau = 2 * np.pi * tau(adv[plane].loc[k1_el, el_out_all], q)
                    brackets = 2 * pi_term + (
                        (np.sin(2 * pi2tau.loc[:, el_out].to_numpy())
                         - np.sin(2 * pi2tau.loc[:, el_out_mm].to_numpy())) / np.sin(2 * np.pi * q)
                    )
                    dmu[plane] = pd.DataFrame(
                        tw.loc[k1_el, col_beta].to_numpy()[:, None] * brackets * (coeff_sign / (8 * np.pi)),
                        index=k1_el,
                        columns=el_out,
                    ).transpose()
            else:
                LOG.debug("  No 'K1L' variables found. Phase Response will be empty.")
                dmu = {"X": pd.DataFrame(None, index=el_out),
                       "Y": pd.DataFrame(None, index=el_out)}

        return dmu

    def _calc_phase_response(self):
        """Response Matrix for delta DPhi.

        This calculation could also be achieved by applying np.cumsum to the DataFrames of
        _calc_phase_adv_response() (tested!), but _calc_phase_response() is about 4x faster.
        """
        LOG.debug("Calculate Phase Response Matrix")
        with timeit(lambda t: LOG.debug(f"  Time needed: {t} s")):
            tw = self._twiss
            adv = self._phase_advances
            k1_el = self._elements_in["K1L"]
            el_out = self._elements_out

            if len(k1_el) > 0:
                dmu = dict.fromkeys(PLANES)

                pi = pd.DataFrame(tw["S"].to_numpy()[:, None] < tw["S"].to_numpy()[None, :],  # pi(i,j) = s(i) < s(j)
                                      index=tw.index, columns=tw.index, dtype=int)

                pi_term = pi.loc[k1_el, el_out].to_numpy()

                for plane in PLANES:
                    col_beta = f"{BETA}{plane}"
                    q = tw.Q1 if plane == "X" else tw.Q2
                    coeff_sign = 1 if plane == "X" else -1

                    pi2tau = 2 * np.pi * tau(adv[plane].loc[k1_el, [DUMMY_ID] + el_out], q)
                    brackets = 2 * pi_term + (
                        (np.sin(2 * pi2tau.loc[:, el_out].to_numpy())
                         - np.sin(2 * pi2tau.loc[:, DUMMY_ID].to_numpy()[:, None])
                        ) / np.sin(2 * np.pi * q)
                    )
                    dmu[plane] = pd.DataFrame(
                        tw.loc[k1_el, col_beta].to_numpy()[:, None] * brackets * (coeff_sign / (8 * np.pi)),
                        index=k1_el,
                        columns=el_out,
                    ).transpose()
            else:
                LOG.debug("  No 'K1L' variables found. Phase Response will be empty.")
                dmu = {"X": pd.DataFrame(None, index=el_out),
                       "Y": pd.DataFrame(None, index=el_out)}

        return dmu

    def _calc_tune_response(self):
        """Response vectors for Tune."""
        LOG.debug("Calculate Tune Response Matrix")
        with timeit(lambda t: LOG.debug(f"  Time needed: {t} s")):
            tw = self._twiss
            k1_el = self._elements_in["K1L"]

            if len(k1_el) > 0:
                dtune = dict.fromkeys(PLANES)

                dtune["X"] = 1 / (4 * np.pi) * tw.loc[k1_el, [f"{BETA}X"]].transpose()
                dtune["X"].index = ["DQX"]

                dtune["Y"] = -1 / (4 * np.pi) * tw.loc[k1_el, [f"{BETA}Y"]].transpose()
                dtune["Y"].index = ["DQY"]
            else:
                LOG.debug("  No 'K1L' variables found. Tune Response will be empty.")
                dtune = {"X": pd.DataFrame(None, index=["DQX"]),
                         "Y": pd.DataFrame(None, index=["DQY"])}
        return dtune

    ################################
    #       Normalizing
    ################################

    def _normalize_beta_response(self, beta):
        """ Convert to Beta Beating """
        el_out = self._elements_out
        tw = self._twiss

        beta_norm = dict.fromkeys(beta.keys())
        for plane in beta:
            col = f"{BETA}{plane}"
            beta_norm[plane] = beta[plane].div(tw.loc[el_out, col], axis="index")
        return beta_norm

    ################################
    #       Mapping
    ################################

    def _map_dispersion_response(self, disp):
        """ Maps all dispersion matrices """
        disp_mapped = dict.fromkeys(disp.keys())
        m2v = self._map_to_variables
        for plane in disp:
            mapping = self._var_to_el[plane.split("_")[1]]
            disp_mapped[plane] = m2v(disp[plane], mapping)
        return disp_mapped

    @staticmethod
    def _map_to_variables(df, mapping):
        r"""Maps from magnets to variables using self._var_to_el.
        Could actually be done by matrix multiplication :math:'A \cdot var_to_el',
         yet, as var_to_el is very sparsely populated, looping is easier.

        Args:
            df: DataFrame or dictionary of DataFrames to map
            mapping: mapping to be applied (e.g. var_to_el[order])
        Returns:
            DataFrame or dictionary of mapped DataFrames
        """

        def map_fun(df, mapping):
            """ Actual mapping function """
            df_map = pd.DataFrame(index=df.index, columns=mapping.keys())
            for var, magnets in mapping.items():
                df_map[var] = df.loc[:, upper(magnets.index)].mul(magnets.to_numpy(), axis="columns").sum(axis="columns")
            return df_map

        # convenience wrapper for dicts
        if isinstance(df, dict):
            mapped = dict.fromkeys(df.keys())
            for plane in mapped:
                mapped[plane] = map_fun(df[plane], mapping)
        else:
            mapped = map_fun(df, mapping)
        return mapped

    ################################
    #          Getters
    ################################

    def get_beta_beat(self, mapped=True):
        """ Returns Response Matrix for Beta Beating """
        if not self._beta:
            self._beta = self._calc_beta_response()
        if not self._beta_beat:
            self._beta_beat = self._normalize_beta_response(self._beta)

        if mapped and not self._beta_beat_mapped:
            self._beta_beat_mapped = self._map_to_variables(self._beta_beat, self._var_to_el["K1L"])
        return self._beta_beat_mapped if mapped else self._beta_beat

    def get_dispersion(self, mapped=True):
        """ Returns Response Matrix for Dispersion """
        if not self._dispersion:
            self._dispersion = self._calc_dispersion_response()

        if mapped and not self._dispersion_mapped:
            self._dispersion_mapped = self._map_dispersion_response(self._dispersion)

        return self._dispersion_mapped if mapped else self._dispersion

    def get_norm_dispersion(self, mapped=True):
        """ Returns Response Matrix for Normalized Dispersion """
        if not self._norm_dispersion:
            self._norm_dispersion = self._calc_norm_dispersion_response()

        if mapped and not self._norm_dispersion_mapped:
            self._norm_dispersion_mapped = self._map_dispersion_response(self._norm_dispersion)
        return self._norm_dispersion_mapped if mapped else self._norm_dispersion

    def get_phase(self, mapped=True):
        """ Returns Response Matrix for Total Phase """
        if not self._phase:
            self._phase = self._calc_phase_response()

        if mapped and not self._phase_mapped:
            self._phase_mapped = self._map_to_variables(self._phase, self._var_to_el["K1L"])

        if mapped:
            return self._phase_mapped
        else:
            return self._phase

    def get_phase_adv(self, mapped=True):
        """ Returns Response Matrix for Phase Advance """
        if not self._phase_adv:
            self._phase_adv = self._calc_phase_advance_response()

        if mapped and not self._phase_adv_mapped:
            self._phase_adv_mapped = self._map_to_variables(self._phase_adv, self._var_to_el["K1L"])

        if mapped:
            return self._phase_adv_mapped
        else:
            return self._phase_adv

    def get_tune(self, mapped=True):
        """ Returns Response Matrix for the Tunes """
        if not self._tune:
            self._tune = self._calc_tune_response()

        if mapped and not self._tune_mapped:
            self._tune_mapped = self._map_to_variables(self._tune, self._var_to_el["K1L"])

        if mapped:
            return self._tune_mapped
        else:
            return self._tune

    def get_coupling(self, mapped=True):
        """ Returns Response Matrix for the coupling """
        if not self._coupling:
            self._coupling = self._calc_coupling_response()

        if mapped and not self._coupling_mapped:
            self._coupling_mapped = self._map_to_variables(self._coupling, self._var_to_el["K1SL"])

        if mapped:
            return self._coupling_mapped
        else:
            return self._coupling

    def get_variable_names(self):
        return self._variables

    def get_variable_mapping(self, order=None):
        if order is None:
            return self._var_to_el
        else:
            return self._var_to_el[order]

    def get_response_for(self, observables=None) -> dict:  # Dict[str, ???]
        """ Calculates and returns only desired response matrices """
        # calling functions for the getters to call functions only if needed
        def caller(func, plane):
            return func()[plane]

        def disp_caller(func, plane):
            disp = func()
            return response_add(*[disp[k] for k in disp.keys() if k.startswith(plane)])

        def tune_caller(func, _unused):
            tune = func()
            res = tune["X"].append(tune["Y"])
            res.index = [f"{TUNE}1", f"{TUNE}2"]
            return res

        def couple_caller(func, plane):
            # apply() converts empty DataFrames to Series! Cast them back.
            # Also: take care of minus-sign convention!
            sign = -1 if plane[-1] == "R" else 1
            part_func = np.real if plane[-1] == "R" else np.imag
            return sign * pd.DataFrame(func()[plane[:-1]].apply(part_func).astype(np.float64))

        # to avoid if-elif-elif-...
        obs_map = {
            f"{TUNE}": (tune_caller, self.get_tune, None),
            f"{BETA}X": (caller, self.get_beta_beat, "X"),
            f"{BETA}Y": (caller, self.get_beta_beat, "Y"),
            f"{PHASE_ADV}X": (caller, self.get_phase, "X"),
            f"{PHASE_ADV}Y": (caller, self.get_phase, "Y"),
            f"{DISP}X": (disp_caller, self.get_dispersion, "X"),
            f"{DISP}Y": (disp_caller, self.get_dispersion, "Y"),
            f"{NORM_DISP}X": (disp_caller, self.get_norm_dispersion, "X"),
            f"{NORM_DISP}Y": (disp_caller, self.get_norm_dispersion, "Y"),
            f"{F1001}R": (couple_caller, self.get_coupling, "1001R"),
            f"{F1001}I": (couple_caller, self.get_coupling, "1001I"),
            f"{F1010}R": (couple_caller, self.get_coupling, "1010R"),
            f"{F1010}I": (couple_caller, self.get_coupling, "1010I"),
        }

        if observables is None:
            observables = list(obs_map.keys())
        else:
            # rename to PHASE to MU
            observables = [f"{PHASE_ADV}{key[-1]}" if key.startswith(PHASE) else key for key in observables]

        LOG.debug(f"Calculating responses for {observables}.")
        with timeit(lambda t: LOG.debug(f"Total time getting responses: {t} s")):
            response = dict.fromkeys(observables)
            for key in observables:
                # response[key] = obs_map[key][0](*obs_map[key][1:3])
                # I still don't know why there is a sign-change between B1 and B2 (jdilly, 2021)
                response[key] = self._beam_sign * obs_map[key][0](*obs_map[key][1:3])
        return response


# Associated Functions #########################################################


def response_add(*args) -> pd.DataFrame:
    """ Merges two or more Response Matrix DataFrames """
    base_df = args[0]
    for df in args[1:]:
        base_df = base_df.add(df, fill_value=0.0)
    return base_df


def upper(list_of_strings: Sequence[str]) -> Sequence[str]:
    """ Set all items of list to uppercase """
    return [item.upper() for item in list_of_strings]


def get_phase_advances(twiss_df: pd.DataFrame) -> Dict[str, pd.DataFrame]:
    """
    Calculate phase advances between all elements

    Returns:
        Matrices similar to DPhi(i,j) = Phi(j) - Phi(i)
    """
    LOG.debug("Calculating Phase Advances:")
    phase_advance_dict = dict.fromkeys(["X", "Y"])
    with timeit(lambda elapsed: LOG.debug(f"  Phase Advances calculated in {elapsed}s")):
        for plane in PLANES:
            colmn_phase = f"{PHASE_ADV}{plane}"
            phases_mdl = twiss_df.loc[twiss_df.index, colmn_phase]
            # Same convention as in [1]: DAdv(i,j) = Phi(j) - Phi(i)
            phase_advances = pd.DataFrame((phases_mdl.to_numpy()[None, :] - phases_mdl.to_numpy()[:, None]),
                                          index=twiss_df.index,
                                          columns=twiss_df.index)
            # Do not calculate dphi and tau here.
            # only slices of phase_advances as otherwise super slow
            phase_advance_dict[plane] = phase_advances
    return phase_advance_dict


def dphi(data, q):
    """Return dphi from phase advances in data, see Eq. 7 in [#DillyUpdatedGlobalOpticsCorrection2018]_"""
    return data + np.where(data <= 0, q, 0)  # '<=' seems to be what MAD-X does


def tau(data, q):
    """Return tau from phase advances in data, see Eq. 8 in [#DillyUpdatedGlobalOpticsCorrection2018]_"""
    return data + np.where(data <= 0, q / 2, -q / 2)  # '<=' seems to be what MAD-X does


# Wrapper ##################################################################


def create_response(
    accel_inst: Accelerator,
    vars_categories: Sequence[str],
    optics_params: List[str],
) -> dict:
    """ Wrapper to create response via TwissResponse """
    LOG.debug("Creating response via TwissResponse.")
    varmap_path = check_varmap_file(accel_inst, vars_categories)

    with timeit(lambda t: LOG.debug(f"Total time getting TwissResponse: {t} s")):
        tr = TwissResponse(accel_inst, vars_categories, varmap_path)
        response: dict = tr.get_response_for(optics_params)

    if not any([resp.size for resp in response.values()]):
        raise ValueError("Responses are all empty. "
                         f"Are variables {tr.get_variable_names()} correct for '{optics_params}'?")
    return response



"""
Handler
-------

This module contains high-level functions to manage most functionality of ``harpy``.
Tools are provided to handle the cleaning, frequency analysis and resonance search for a
single-bunch `TbtData`.
"""
from collections import OrderedDict
from os.path import basename, join

import numpy as np
import pandas as pd
import tfs

from omc3.definitions import formats
from omc3.definitions.constants import PLANES, PLANE_TO_NUM as P2N
from omc3.harpy import clean, frequency, kicker
from omc3.harpy.constants import (FILE_AMPS_EXT, FILE_FREQS_EXT, FILE_LIN_EXT,
                                  COL_NAME, COL_TUNE, COL_AMP, COL_MU,
                                  COL_NATTUNE, COL_NATAMP, COL_PHASE, COL_ERR)
from omc3.utils import logging_tools
from omc3.utils.contexts import timeit

LOGGER = logging_tools.get_logger(__name__)
ALL_PLANES = (*PLANES, "Z")
PLANE_TO_NUM = {**P2N, "Z": 3}


def run_per_bunch(tbt_data, harpy_input):
    """
    Cleans data, analyses frequencies and searches for resonances.

    Args:
        tbt_data: single bunch `TbtData`.
        harpy_input: Analysis settings taken from the commandline.

    Returns:
        Dictionary with a `TfsDataFrame` per plane.
    """
    model = None if harpy_input.model is None else tfs.read(harpy_input.model, index=COL_NAME).loc[:, 'S']
    bpm_datas, usvs, lins, bad_bpms = {}, {}, {}, {}
    output_file_path = _get_output_path_without_suffix(harpy_input.outputdir, harpy_input.files)
    for plane in PLANES:
        bpm_data = _get_cut_tbt_matrix(tbt_data, harpy_input.turns, plane)
        bpm_data = _scale_to_meters(bpm_data, harpy_input.unit)
        bpm_data, usvs[plane], bad_bpms[plane], bpm_res = clean.clean(harpy_input, bpm_data, model)
        lins[plane], bpm_datas[plane] = _closed_orbit_analysis(bpm_data, model, bpm_res)

    tune_estimates = harpy_input.tunes if harpy_input.autotunes is None else frequency.estimate_tunes(
        harpy_input, usvs if harpy_input.clean else
        dict(X=clean.svd_decomposition(bpm_datas["X"], harpy_input.sing_val),
             Y=clean.svd_decomposition(bpm_datas["Y"], harpy_input.sing_val)))

    spectra = {}
    for plane in PLANES:
        with timeit(lambda spanned: LOGGER.debug(f"Time for harmonic_analysis: {spanned}")):
            harpy_results, spectra[plane], bad_bpms_summaries = frequency.harpy_per_plane(
                harpy_input, bpm_datas[plane], usvs[plane], tune_estimates, plane)
        if "bpm_summary" in harpy_input.to_write:
            bad_bpms[plane].extend(bad_bpms_summaries)
            _write_bad_bpms(output_file_path, plane, bad_bpms[plane])
        if "spectra" in harpy_input.to_write or "full_spectra" in harpy_input.to_write:
            _write_spectrum(output_file_path, plane, spectra[plane])
        lins[plane] = lins[plane].loc[harpy_results.index].join(harpy_results)
        if harpy_input.is_free_kick:
            lins[plane] = kicker.phase_correction(bpm_datas[plane], lins[plane], plane)

    measured_tunes = [lins["X"][f"{COL_TUNE}X"].mean(), lins["Y"][f"{COL_TUNE}Y"].mean(),
                      lins["X"][f"{COL_TUNE}Z"].mean() if tune_estimates[2] > 0 else 0]

    for plane in PLANES:
        lins[plane] = lins[plane].join(frequency.find_resonances(
            measured_tunes, bpm_datas[plane].shape[1], plane, spectra[plane], harpy_input.resonances))
        lins[plane] = _add_calculated_phase_errors(lins[plane])
        lins[plane] = _sync_phase(lins[plane], plane)
        lins[plane] = _rescale_amps_to_main_line_and_compute_noise(lins[plane], plane)
        lins[plane] = lins[plane].sort_values('S', axis=0, ascending=True)
        lins[plane] = tfs.TfsDataFrame(lins[plane], headers=_compute_headers(lins[plane], tbt_data.date))
        if "lin" in harpy_input.to_write:
            _write_lin_tfs(output_file_path, plane, lins[plane])
    return lins


def _get_cut_tbt_matrix(tbt_data, turn_indices, plane):
    start = max(0, min(turn_indices))
    end = min(max(turn_indices), tbt_data.matrices[0][plane].shape[1])
    return tbt_data.matrices[0][plane].iloc[:, start:end].T.reset_index(drop=True).T


def _scale_to_meters(bpm_data, unit):
    scales_to_meters = {'um': 1e-6, 'mm': 0.001, 'cm': 0.01, 'm': 1}
    bpm_data.iloc[:,:] = bpm_data.iloc[:,:].to_numpy() * scales_to_meters[unit]
    return bpm_data


def _closed_orbit_analysis(bpm_data, model, bpm_res):
    lin_frame = pd.DataFrame(index=bpm_data.index.to_numpy(),
                             data=OrderedDict([(COL_NAME, bpm_data.index.to_numpy()),
                                               ("S", np.arange(bpm_data.index.size) if model is None
                                               else model.loc[bpm_data.index])]))
    lin_frame['BPM_RES'] = 0.0 if bpm_res is None else bpm_res.loc[lin_frame.index]
    with timeit(lambda spanned: LOGGER.debug(f"Time for orbit_analysis: {spanned}")):
        lin_frame = _get_orbit_data(lin_frame, bpm_data)
    return lin_frame, bpm_data.subtract(bpm_data.mean(axis=1), axis=0)


def _get_orbit_data(lin_frame, bpm_data):
    lin_frame['PK2PK'] = np.max(bpm_data, axis=1) - np.min(bpm_data, axis=1)
    lin_frame['CO'] = np.mean(bpm_data, axis=1)
    lin_frame['CORMS'] = np.std(bpm_data, axis=1) / np.sqrt(bpm_data.shape[1])
    # TODO: Magic number 10?: Maybe accelerator dependent ... LHC 6-7?
    lin_frame['NOISE'] = lin_frame.loc[:, 'BPM_RES'] / np.sqrt(bpm_data.shape[1]) / 10.0
    return lin_frame


def _add_calculated_phase_errors(lin_frame):
    noise = lin_frame.loc[:, 'NOISE'].to_numpy()
    if np.max(noise) == 0.0:
        return lin_frame   # Do not calculated errors when no noise was calculated
    for name_root in (COL_MU, COL_PHASE):
        cols = [col for col in lin_frame.columns.to_numpy() if name_root in col]
        for col in cols:
            lin_frame[f"{COL_ERR}{col}"] = _get_spectral_phase_error(
                lin_frame.loc[:, f"{col.replace(name_root, COL_AMP)}"], noise)
    return lin_frame


def _get_spectral_phase_error(amplitude, noise):
    """
    When the error is too big (> 2*pi*0.25 more or less) the noise is not Gaussian anymore.
    In such a case the distribution is almost uniform, so we set the error to be 0.3, which is
    the standard deviation of uniformly distributed phases.
    This approximation does not bias the error by more than 20%, and that is only for large errors.
    """
    error = noise / (np.where(amplitude > 0.0, amplitude, 1e-15) * 2 * np.pi)
    return np.where(error > 0.25, 0.3, error)


def _sync_phase(lin_frame, plane):
    """
    Produces ``MUXSYNC`` and ``MUYSYNC`` columns that are ``MUX/MUY`` but shifted such that for
    BPM at index 0 is always 0. It allows to compare phases of consecutive measurements and if
    some measurements stick out remove them from the data set. Original author is **skowron**.
    """
    phase = lin_frame.loc[:, f"{COL_MU}{plane}"].to_numpy()
    phase = phase - phase[0]
    lin_frame[f"{COL_MU}{plane}SYNC"] = np.where(np.abs(phase) > 0.5, phase - np.sign(phase), phase)
    return lin_frame


def _compute_headers(panda, date=None):
    headers = OrderedDict()
    for plane in ALL_PLANES:
        for prefix in ("", "NAT"):
            try:
                bpm_tunes = panda.loc[:, f"{prefix}{COL_TUNE}{plane}"]
            except KeyError:
                pass
            else:
                headers[f"{prefix}Q{PLANE_TO_NUM[plane]}"] = np.mean(bpm_tunes)
                headers[f"{prefix}Q{PLANE_TO_NUM[plane]}RMS"] = np.std(bpm_tunes)
    if date:
        headers["TIME"] = date.strftime(formats.TIME)
    return headers


def _write_bad_bpms(output_path_without_suffix, plane, bad_bpms_with_reasons):
    with open(f"{output_path_without_suffix}.bad_bpms_{plane.lower()}", 'w') as bad_bpms_file:
        for line in bad_bpms_with_reasons:
            bad_bpms_file.write(f"{line}\n")


def _write_spectrum(output_path_without_suffix, plane, spectra):
    tfs.write(f"{output_path_without_suffix}{FILE_AMPS_EXT.format(plane=plane.lower())}", spectra["COEFFS"].abs().T)
    tfs.write(f"{output_path_without_suffix}{FILE_FREQS_EXT.format(plane=plane.lower())}", spectra["FREQS"].T)


def _write_lin_tfs(output_path_without_suffix, plane, lin_frame):
    tfs.write(f"{output_path_without_suffix}{FILE_LIN_EXT.format(plane=plane.lower())}", lin_frame)


def _get_output_path_without_suffix(output_dir, file_path):
    return join(output_dir, basename(file_path))


def _rescale_amps_to_main_line_and_compute_noise(panda, plane):
    """
    TODO    follows non-transpararent convention
    TODO    the consequent analysis has to be changed if removed
    """
    cols = [col for col in panda.columns.to_numpy() if col.startswith(COL_AMP)]
    cols.remove(f"{COL_AMP}{plane}")
    panda.loc[:, cols] = panda.loc[:, cols].div(panda.loc[:, f"{COL_AMP}{plane}"], axis="index")
    amps = panda.loc[:, f"{COL_AMP}{plane}"].to_numpy()
    # Division by two for backwards compatibility with Drive, i.e. the unit is [2mm]
    # TODO  later remove
    panda[f"{COL_AMP}{plane}"] = panda.loc[:, f"{COL_AMP}{plane}"].to_numpy() / 2
    if f"{COL_NATAMP}{plane}" in panda.columns:
        panda[f"{COL_NATAMP}{plane}"] = panda.loc[:, f"{COL_NATAMP}{plane}"].to_numpy() / 2

    if np.max(panda.loc[:, 'NOISE'].to_numpy()) == 0.0:
        return panda  # Do not calculated errors when no noise was calculated
    noise_scaled = panda.loc[:, 'NOISE'] / amps
    panda.loc[:, "NOISE_SCALED"] = noise_scaled
    panda.loc[:, f"{COL_ERR}{COL_AMP}{plane}"] = panda.loc[:, 'NOISE']
    if f"{COL_NATTUNE}{plane}" in panda.columns:
        panda.loc[:, f"{COL_ERR}{COL_NATAMP}{plane}"] = panda.loc[:, 'NOISE']
    for col in cols:
        this_amp = panda.loc[:, col]
        panda.loc[:, f"{COL_ERR}{col}"] = noise_scaled * np.sqrt(1 + np.square(this_amp))
    return panda



"""
Formats
-------------

Recurring formats are defined here.
"""
from pathlib import Path
from datetime import datetime

TIME = "%Y_%m_%d@%H_%M_%S_%f"  # CERN default
CONFIG_FILENAME = "{script:s}_{time:s}.ini"
BACKUP_FILENAME = "{basefile}_backup{counter}"


def get_config_filename(script):
    """Default Filename for config-files. Call from script with ``__file__``."""
    return CONFIG_FILENAME.format(
        script=Path(script).name.split('.')[0],
        time=datetime.utcnow().strftime(TIME)
    )




"""
Logging Tools
-------------

Functions for easier use of logging, like automatic logger setup
(see: :meth:`~utils.logging_tools.get_logger`).
"""
import datetime
import inspect
import logging
import os
import sys
import time
import warnings
from contextlib import contextmanager
from io import StringIO
from logging import NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL  # make them available directly

import numpy as np
import pandas as pd

DIVIDER = "|"
NEWLINE = "\n" + " " * 10  # levelname + divider + 2
BASIC_FORMAT = '%(levelname)7s {div:s} %(message)s {div:s} %(name)s'.format(div=DIVIDER)
COLOR_LEVEL = '\33[0m\33[38;2;150;150;255m'
COLOR_MESSAGE = '\33[0m'
COLOR_MESSAGE_LOW = '\33[0m\33[38;2;140;140;140m'
COLOR_WARN = '\33[0m\33[38;2;255;161;53m'
COLOR_ERROR = '\33[0m\33[38;2;216;31;42m'
COLOR_NAME = '\33[0m\33[38;2;80;80;80m'
COLOR_DIVIDER = '\33[0m\33[38;2;127;127;127m'
COLOR_RESET = '\33[0m'

MADX = DEBUG + 3

# Classes and Contexts #########################################################


class MaxFilter(object):
    """To get messages only up to a certain level."""
    def __init__(self, level):
        self.__level = level

    def filter(self, log_record):
        return log_record.levelno <= self.__level


class DebugMode(object):
    """
    Context Manager for the debug mode.
    Hint: Does not work with ``@contextmanager`` from contextlib (even though nicer code),
    as the ``_get_caller`` would find **contextlib.py**.

    Args:
        active (bool): Defines if this manager is doing anything. Defaults to ``True``.
        log_file (str): File to log into.
    """
    def __init__(self, active=True, log_file=None):
        self.active = active
        if active:
            # get current logger
            caller_file = _get_caller()
            current_module = _get_current_module(caller_file)

            self.logger = logging.getLogger(".".join([current_module, os.path.basename(caller_file)]))

            # set level to debug
            self.current_level = self.logger.getEffectiveLevel()
            self.logger.setLevel(DEBUG)
            self.logger.debug("Running in Debug-Mode.")

            # create logfile name:
            now = "{:s}_".format(datetime.datetime.now().isoformat())
            if log_file is None:
                log_file = os.path.abspath(caller_file).replace(".pyc", "").replace(".py",
                                                                                    "") + ".log"
            self.log_file = os.path.join(os.path.dirname(log_file), now + os.path.basename(log_file))
            self.logger.debug("Writing log to file '{:s}'.".format(self.log_file))

            # add handlers
            self.file_h = file_handler(self.log_file, level=DEBUG)
            self.console_h = stream_handler(level=DEBUG, max_level=DEBUG)
            self.mod_logger = logging.getLogger(current_module)
            self.mod_logger.addHandler(self.file_h)
            self.mod_logger.addHandler(self.console_h)

            # stop time
            self.start_time = time.time()

    def __enter__(self):
        return None

    def __exit__(self, *args, **kwargs):
        if self.active:
            # summarize
            time_used = time.time() - self.start_time
            log_id = "" if self.log_file is None else "'{:s}'".format(
                os.path.basename(self.log_file))
            self.logger.debug("Exiting Debug-Mode {:s} after {:f}s.".format(log_id, time_used))

            # revert everything
            self.logger.setLevel(self.current_level)
            self.mod_logger.removeHandler(self.file_h)
            self.mod_logger.removeHandler(self.console_h)


class TempFile(object):
    """
    Context Manager. Lets another function write into a temporary file and logs its contents.
    It won't open the file, so only the files path is returned.

    Args:
        file_path (str): Place to write the tempfile to.
        log_func (func): The function with which the content should be logged (e.g. LOG.info).
    """

    def __init__(self, file_path, log_func):
        self.path = file_path
        self.log_func = log_func

    def __enter__(self):
        return self.path

    def __exit__(self, value, traceback):
        try:
            with open(self.path, "r") as f:
                content = f.read()
            self.log_func("{:s}:\n".format(self.path) + content)
        except IOError:
            self.log_func("{:s}: -file does not exist-".format(self.path))
        else:
            os.remove(self.path)


@contextmanager
def log_pandas_settings_with_copy(log_func):
    """Logs pandas ``SettingsWithCopy`` warning to loc_func instead of printing the warning."""
    caller_line = inspect.currentframe().f_back.f_back.f_lineno  # one frame for contextmanager
    old_mode = pd.options.mode.chained_assignment
    pd.options.mode.chained_assignment = 'warn'
    try:
        with warnings.catch_warnings(record=True) as warn_list:
            yield
        for w in warn_list:
            if not issubclass(w.category, pd.core.common.SettingWithCopyWarning):
                warnings.warn(w)
            else:
                message = w.message.args[0].split("\n")
                log_func("{:s} (l. {:d})".format(message[1], caller_line))
    finally:
        pd.options.mode.chained_assignment = old_mode


@contextmanager
def logging_silence():
    """Remove temporarily all loggers from root logger."""
    root_logger = getLogger("")
    handlers = list(root_logger.handlers)
    root_logger.handlers = []

    yield

    root_logger.handlers = handlers


@contextmanager
def unformatted_console_logging():
    """Log only to console and only unformatted."""
    with logging_silence():
        handler = stream_handler(level=NOTSET, fmt="%(message)s")
        rl = getLogger("")
        rl.addHandler(handler)

        yield

        rl.removeHandler(handler)


class TempStringLogger:
    """
    Temporarily log into a string that can be retrieved by ``get_log``.

    Args:
        module: module to log, defaults to the caller file.
        level: logging level, defaults to ``INFO``.
    """
    def __init__(self, module=None, level=INFO):
        if module is None:
            module = _get_caller_logger_name()

        self.stream = StringIO()
        self.level = level
        self.handler = stream_handler(stream=self.stream)
        self.log = getLogger(module)

    def __enter__(self):
        self.log.propagate = False
        self.log.setLevel(self.level)
        self.log.addHandler(self.handler)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.log.removeHandler(self.handler)

    def get_log(self):
        """ Get the log as string. """
        return self.stream.getvalue()


def odr_pprint(printer, odr_out):
    """Logs the odr output results. Adapted from odr_output pretty print."""
    old_opts = np.get_printoptions()
    np.set_printoptions(precision=2)
    covariance_str = str(odr_out.cov_beta).replace("\n", ",")
    odr_str = ('\nODR-Summary:\n'
               f'  Beta: {odr_out.beta}\n'
               f'  Beta Std Error: {odr_out.sd_beta}\n'
               f'  Beta Covariance: {covariance_str}\n'
               )
    if hasattr(odr_out, 'info'):
        odr_str += (f'  Residual Variance: {odr_out.res_var:.2e}\n'
                    f'  Inverse Condition #: {odr_out.inv_condnum:.2e}\n'
                    f'  Reason(s) for Halting:\n')
        for r in odr_out.stopreason:
            odr_str += f'    {r}\n'
    printer(odr_str)
    np.set_printoptions(**old_opts)


def list2str(list_: list) -> str:
    """Returns string representation of ``list_``, but without brackets."""
    return str(list_).lstrip("[").rstrip("]")

# Public Methods ###############################################################


def get_logger(name, level_root=DEBUG, level_console=None, fmt=BASIC_FORMAT, color=None):
    """
    Sets up logger if name is **__main__**. Returns logger based on module name.

    Args:
        name: only used to check if __name__ is __main__.
        level_root: main logging level, defaults to ``DEBUG``.
        level_console: console logging level, defaults to ``INFO``.
        fmt: Format of the logging. For default see ``BASIC_FORMAT``.
        color: If `None` colors are used if tty is detected.
              `False` will never use colors and `True` will always enforce them.

    Returns:
        Logger instance.
    """
    logger_name = _get_caller_logger_name()

    if name == "__main__":
        if level_console is None:
            level_console = DEBUG if sys.flags.debug else INFO

        # set up root logger
        root_logger = logging.getLogger("")
        root_logger.handlers = []  # remove handlers in case someone already created them
        root_logger.setLevel(level_root)

        logging.addLevelName(MADX, 'MADX')

        # print logs to the console
        root_logger.addHandler(
            stream_handler(
                level=max(level_console, DEBUG),
                max_level=INFO-1,
                fmt=_maybe_bring_color(fmt, DEBUG, color),
            )
        )

        root_logger.addHandler(
            stream_handler(
                level=max(level_console, INFO),
                max_level=WARNING-1,
                fmt=_maybe_bring_color(fmt, INFO, color),
            )
        )

        # print console warnings
        root_logger.addHandler(
            stream_handler(
                level=max(WARNING, level_console),
                max_level=ERROR-1,
                fmt=_maybe_bring_color(fmt, WARNING, color),
            )
        )

        # print errors to error-stream
        root_logger.addHandler(
            stream_handler(
                stream=sys.stderr,
                level=max(ERROR, level_console),
                fmt=_maybe_bring_color(fmt, ERROR, color),
            )
        )

    # logger for the current file
    return logging.getLogger(logger_name)


def file_handler(logfile, level=DEBUG, fmt=BASIC_FORMAT):
    """Convenience function so the caller does not have to import logging."""
    handler = logging.FileHandler(logfile, mode='w', )
    handler.setLevel(level)
    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)
    return handler


def stream_handler(stream=sys.stdout, level=DEBUG, fmt=BASIC_FORMAT, max_level=None):
    """Convenience function so the caller does not have to import logging."""
    handler = logging.StreamHandler(stream)
    handler.setLevel(level)
    console_formatter = logging.Formatter(fmt)
    handler.setFormatter(console_formatter)
    if max_level:
        handler.addFilter(MaxFilter(max_level))
    return handler


def add_module_handler(handler):
    """Add handler at current module level."""
    current_module = _get_current_module()
    logging.getLogger(current_module).addHandler(handler)


def add_root_handler(handler):
    """Add handler at root level."""
    logging.getLogger("").addHandler(handler)


def getLogger(name):
    """Convenience function so the caller does not have to import logging."""
    return logging.getLogger(name)


def get_my_logger_name():
    """Return the logger name for the caller."""
    return _get_caller_logger_name()


# Private Methods ##############################################################


def _get_caller():
    """Find the caller of the current log-function."""
    this_file, _ = os.path.splitext(__file__)
    caller_file = this_file
    caller_frame = inspect.currentframe()
    while this_file == caller_file:
        caller_frame = caller_frame.f_back
        (caller_file_full, _, _, _, _) = inspect.getframeinfo(caller_frame)
        caller_file, _ = os.path.splitext(caller_file_full)
    return caller_file


def _get_current_module(current_file=None):
    """Find the name of the current module."""
    if not current_file:
        current_file = _get_caller()
    path_parts = os.path.abspath(current_file).split(os.path.sep)

    repo_parts = os.path.abspath(
                    os.path.join(os.path.dirname(os.path.abspath(__file__)), os.path.pardir)
                    ).split(os.path.sep)

    current_module = '.'.join(path_parts[len(repo_parts):-1])
    return current_module


def _get_caller_logger_name():
    """Returns logger name of the caller."""
    caller_file = _get_caller()
    current_module = _get_current_module(caller_file)
    return ".".join([current_module, os.path.basename(caller_file)])


def _maybe_bring_color(format_string, colorlevel=INFO, color_flag=None):
    """Adds color to the logs (can only be used in a terminal)."""
    if color_flag is None:
        color_flag = _isatty()

    if not color_flag:
        return format_string

    level = "%(levelname)"
    message = "%(message)"
    name = "%(name)"

    if colorlevel <= WARNING:
        format_string = format_string.replace(level, COLOR_LEVEL + level)
    else:
        format_string = format_string.replace(level, COLOR_ERROR + level)

    if colorlevel <= DEBUG:
        format_string = format_string.replace(message, COLOR_MESSAGE_LOW + message)
    elif colorlevel <= INFO:
        format_string = format_string.replace(message, COLOR_MESSAGE + message)
    elif colorlevel <= WARNING:
        format_string = format_string.replace(message, COLOR_WARN + message)
    else:
        format_string = format_string.replace(message, COLOR_ERROR + message)

    format_string = format_string.replace(name, COLOR_NAME + name)
    format_string = format_string.replace(DIVIDER, COLOR_DIVIDER + DIVIDER)
    format_string = format_string + COLOR_RESET

    return format_string


def _isatty():
    """Checks if stdout is a tty, which means it should support color-codes."""
    return hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
